"""Unit Test for otx.algorithms.detection.adapters.mmdet.task."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import json
from contextlib import nullcontext
from copy import deepcopy
from typing import Any, Dict

import numpy as np
import pytest
import torch
from mmcv.utils import Config
from torch import nn

from otx.algorithms.common.adapters.mmcv.utils.config_utils import MPAConfig
from otx.algorithms.classification.adapters.mmcls.task import MMClassificationTask
from otx.algorithms.classification.adapters.mmcls.models.classifiers.sam_classifier import SAMImageClassifier
from otx.algorithms.classification.configs.base import ClassificationConfig
from otx.api.configuration import ConfigurableParameters
from otx.api.configuration.helper import create
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.explain_parameters import ExplainParameters
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.model import (
    ModelConfiguration,
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
)
from otx.api.entities.model_template import InstantiationType, parse_model_template, TaskFamily, TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.classification.test_helper import (
    DEFAULT_CLS_TEMPLATE_DIR,
    init_environment,
    generate_cls_dataset,
    generate_label_schema,
)


class MockModule(nn.Module):
    """Mock class for nn.Module."""

    def forward(self, inputs: Any):
        return inputs


class MockModel(nn.Module):
    """Mock class for pytorch model."""

    def __init__(self):
        super().__init__()
        self.module = MockModule()
        self.module.backbone = MockModule()
        self.backbone = MockModule()

    def forward(self, *args, **kwargs):
        forward_hooks = list(self.module.backbone._forward_hooks.values())
        for hook in forward_hooks:
            hook(1, 2, 3)
        return np.array([[0.3, 0.7]])

    @staticmethod
    def named_parameters(*args, **kwargs):
        return {"name": torch.Tensor([0.5])}.items()


class MockDataset(DatasetEntity):
    """Mock class for mm_dataset."""

    def __init__(self, dataset: DatasetEntity):
        self.dataset = dataset
        self.CLASSES = ["1", "2", "3"]

    def __len__(self):
        return len(self.dataset)

    def evaluate(self, prediction, *args, **kwargs):
        return {"mAP": 1.0}


class MockDataLoader:
    """Mock class for data loader."""

    def __init__(self, dataset: DatasetEntity):
        self.dataset = dataset
        self.iter = iter(self.dataset)

    def __len__(self) -> int:
        return len(self.dataset)

    def __next__(self) -> Dict[str, DatasetItemEntity]:
        return {"imgs": next(self.iter)}

    def __iter__(self):
        return self


class MockExporter:
    """Mock class for Exporter."""

    def __init__(self, task):
        self._output_path = task._output_path

    def run(self, *args, **kwargs):
        with open(os.path.join(self._output_path, "openvino.bin"), "wb") as f:
            f.write(np.ndarray([0]))
        with open(os.path.join(self._output_path, "openvino.xml"), "wb") as f:
            f.write(np.ndarray([0]))
        with open(os.path.join(self._output_path, "model.onnx"), "wb") as f:
            f.write(np.ndarray([0]))

        return {
            "outputs": {
                "bin": os.path.join(self._output_path, "openvino.bin"),
                "xml": os.path.join(self._output_path, "openvino.xml"),
                "onnx": os.path.join(self._output_path, "model.onnx"),
            }
        }


class TestMMClassificationTask:
    """Test class for MMClassificationTask.

    Details are explained in each test function.
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        model_template = parse_model_template(os.path.join(DEFAULT_CLS_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)

        mc_task_env, self.mc_cls_dataset = init_environment(hyper_parameters, model_template, False, False, 100)
        self.mc_cls_task = MMClassificationTask(mc_task_env)
        self.mc_cls_label_schema = generate_label_schema(self.mc_cls_dataset.get_labels(), False, False)

        ml_task_env, self.ml_cls_dataset = init_environment(hyper_parameters, model_template, True, False, 100)
        self.ml_cls_task = MMClassificationTask(ml_task_env)
        self.ml_cls_label_schema = generate_label_schema(self.ml_cls_dataset.get_labels(), False, False)

        hl_task_env, self.hl_cls_dataset = init_environment(hyper_parameters, model_template, False, True, 100)
        self.hl_cls_task = MMClassificationTask(hl_task_env)
        self.hl_cls_label_schema = generate_label_schema(self.hl_cls_dataset.get_labels(), False, False)

    @e2e_pytest_unit
    def test_build_model(self, mocker) -> None:
        """Test build_model function."""
        _mock_recipe_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_CLS_TEMPLATE_DIR, "model.py"))
        _mock_recipe_cfg.model.pop("task")
        model = self.mc_cls_task.build_model(_mock_recipe_cfg, True)
        assert isinstance(model, SAMImageClassifier)

    @e2e_pytest_unit
    def test_train_multiclass(self, mocker) -> None:
        """Test train function."""

        def _mock_train_model(*args, **kwargs):
            with open(os.path.join(self.mc_cls_task._output_path, "latest.pth"), "wb") as f:
                torch.save({"dummy": torch.randn(1, 3, 3, 3)}, f)

        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataset",
            return_value=MockDataset(self.mc_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataloader",
            return_value=MockDataLoader(self.mc_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.train_model",
            side_effect=_mock_train_model,
        )
        mocker.patch.object(MMClassificationTask, "build_model", return_value=MockModel())
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_data_parallel",
            return_value=MockModel(),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.FeatureVectorHook",
            return_value=nullcontext(),
        )

        _config = ModelConfiguration(ClassificationConfig("header"), self.mc_cls_label_schema)
        output_model = ModelEntity(self.mc_cls_dataset, _config)
        self.mc_cls_task.train(self.mc_cls_dataset, output_model)
        output_model.performance == 1.0

    @e2e_pytest_unit
    def test_train_multilabel(self, mocker) -> None:
        """Test train function."""

        def _mock_train_model(*args, **kwargs):
            with open(os.path.join(self.ml_cls_task._output_path, "latest.pth"), "wb") as f:
                torch.save({"dummy": torch.randn(1, 3, 3, 3)}, f)

        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataset",
            return_value=MockDataset(self.ml_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataloader",
            return_value=MockDataLoader(self.ml_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.train_model",
            side_effect=_mock_train_model,
        )
        mocker.patch.object(MMClassificationTask, "build_model", return_value=MockModel())
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_data_parallel",
            return_value=MockModel(),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.FeatureVectorHook",
            return_value=nullcontext(),
        )

        _config = ModelConfiguration(ClassificationConfig("header"), self.ml_cls_label_schema)
        output_model = ModelEntity(self.ml_cls_dataset, _config)
        self.ml_cls_task.train(self.ml_cls_dataset, output_model)
        output_model.performance == 1.0

    @e2e_pytest_unit
    def test_train_hierarchicallabel(self, mocker) -> None:
        """Test train function."""

        def _mock_train_model(*args, **kwargs):
            with open(os.path.join(self.hl_cls_task._output_path, "latest.pth"), "wb") as f:
                torch.save({"dummy": torch.randn(1, 3, 3, 3)}, f)

        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataset",
            return_value=MockDataset(self.hl_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataloader",
            return_value=MockDataLoader(self.hl_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.train_model",
            side_effect=_mock_train_model,
        )
        mocker.patch.object(MMClassificationTask, "build_model", return_value=MockModel())
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_data_parallel",
            return_value=MockModel(),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.FeatureVectorHook",
            return_value=nullcontext(),
        )

        _config = ModelConfiguration(ClassificationConfig("header"), self.hl_cls_label_schema)
        output_model = ModelEntity(self.hl_cls_dataset, _config)
        self.hl_cls_task.train(self.hl_cls_dataset, output_model)
        output_model.performance == 1.0

    @e2e_pytest_unit
    def test_infer_multiclass(self, mocker) -> None:
        """Test infer function."""

        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataset",
            return_value=MockDataset(self.mc_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataloader",
            return_value=MockDataLoader(self.mc_cls_dataset),
        )
        mocker.patch.object(MMClassificationTask, "build_model", return_value=MockModel())
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_data_parallel",
            return_value=MockModel(),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.FeatureVectorHook",
            return_value=nullcontext(),
        )

        inference_parameters = InferenceParameters(is_evaluation=True)
        outputs = self.mc_cls_task.infer(self.mc_cls_dataset.with_empty_annotations(), inference_parameters)
        for output in outputs:
            assert output.get_annotations()[-1].get_labels()[0].probability == 0.7

    @e2e_pytest_unit
    def test_infer_multilabel(self, mocker) -> None:
        """Test infer function."""

        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataset",
            return_value=MockDataset(self.ml_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataloader",
            return_value=MockDataLoader(self.ml_cls_dataset),
        )
        mocker.patch.object(MMClassificationTask, "build_model", return_value=MockModel())
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_data_parallel",
            return_value=MockModel(),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.FeatureVectorHook",
            return_value=nullcontext(),
        )

        inference_parameters = InferenceParameters(is_evaluation=True)
        outputs = self.ml_cls_task.infer(self.ml_cls_dataset.with_empty_annotations(), inference_parameters)
        for output in outputs:
            assert output.get_annotations()[-1].get_labels()[0].probability == 0.7

    @e2e_pytest_unit
    def test_infer_hierarchicallabel(self, mocker) -> None:
        """Test infer function."""

        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataset",
            return_value=MockDataset(self.hl_cls_dataset),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_dataloader",
            return_value=MockDataLoader(self.hl_cls_dataset),
        )
        mocker.patch.object(MMClassificationTask, "build_model", return_value=MockModel())
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.build_data_parallel",
            return_value=MockModel(),
        )
        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.FeatureVectorHook",
            return_value=nullcontext(),
        )

        inference_parameters = InferenceParameters(is_evaluation=True)
        outputs = self.hl_cls_task.infer(self.hl_cls_dataset.with_empty_annotations(), inference_parameters)
        for output in outputs:
            assert output.get_annotations()[-1].get_labels()[0].probability == 0.7

    @e2e_pytest_unit
    def test_cls_evaluate(self) -> None:
        """Test evaluate function for classification."""

        _config = ModelConfiguration(ClassificationConfig("header"), self.mc_cls_label_schema)
        _model = ModelEntity(self.mc_cls_dataset, _config)
        resultset = ResultSetEntity(_model, self.mc_cls_dataset, self.mc_cls_dataset)
        self.mc_cls_task.evaluate(resultset)
        assert resultset.performance.score.value == 1.0

    @e2e_pytest_unit
    def test_cls_evaluate_with_empty_annotations(self) -> None:
        """Test evaluate function for classification with empty predictions."""

        _config = ModelConfiguration(ClassificationConfig("header"), self.mc_cls_label_schema)
        _model = ModelEntity(self.mc_cls_dataset, _config)
        resultset = ResultSetEntity(_model, self.mc_cls_dataset, self.mc_cls_dataset.with_empty_annotations())
        self.mc_cls_task.evaluate(resultset)
        assert resultset.performance.score.value == 0.0

    @pytest.mark.parametrize("precision", [ModelPrecision.FP16, ModelPrecision.FP32])
    @e2e_pytest_unit
    def test_export(self, mocker, precision: ModelPrecision) -> None:
        """Test export function.

        <Steps>
            1. Create model entity
            2. Run export function
            3. Check output model attributes
        """
        _config = ModelConfiguration(ClassificationConfig("header"), self.mc_cls_label_schema)
        _model = ModelEntity(self.mc_cls_dataset, _config)

        mocker.patch(
            "otx.algorithms.classification.adapters.mmcls.task.ClassificationExporter",
            return_value=MockExporter(self.mc_cls_task),
        )
        mocker.patch(
            "otx.algorithms.classification.task.embed_ir_model_data",
            return_value=True,
        )

        self.mc_cls_task.export(ExportType.OPENVINO, _model, precision, False)

        assert _model.model_format == ModelFormat.OPENVINO
        assert _model.optimization_type == ModelOptimizationType.MO
        assert _model.precision[0] == precision
        assert _model.get_data("openvino.bin") is not None
        assert _model.get_data("openvino.xml") is not None
        assert _model.precision == self.mc_cls_task._precision
        assert _model.optimization_methods == self.mc_cls_task._optimization_methods
        assert _model.get_data("label_schema.json") is not None

    @e2e_pytest_unit
    def test_explain(self, mocker):
        """Test explain function."""
        explain_parameters = ExplainParameters(
            explainer="ClassWiseSaliencyMap",
            process_saliency_maps=False,
            explain_predicted_classes=True,
        )
        outputs = self.mc_cls_task.explain(self.mc_cls_dataset, explain_parameters)
        assert isinstance(outputs, DatasetEntity)
        assert len(outputs) == 200



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import pytest
import torch

from otx.algorithms.classification.adapters.mmcls.models.heads.semisl_multilabel_cls_head import (
    SemiLinearMultilabelClsHead,
    SemiNonLinearMultilabelClsHead,
)
from otx.algorithms.classification.adapters.mmcls.models.losses.asymmetric_loss_with_ignore import (
    AsymmetricLossWithIgnore,
)
from otx.algorithms.classification.adapters.mmcls.models.losses.barlowtwins_loss import (
    BarlowTwinsLoss,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSemiLinearMultilabelClsHead:
    @pytest.fixture(autouse=True)
    def head_type(self):
        return SemiLinearMultilabelClsHead

    @pytest.fixture(autouse=True)
    def setup(self, head_type) -> None:
        self.num_classes = 2
        self.head_dim = 5
        self.loss = dict(type=AsymmetricLossWithIgnore.__name__, reduction="sum")
        self.aux_loss = dict(type=BarlowTwinsLoss.__name__, off_diag_penality=1.0 / 128.0, loss_weight=1.0)
        self.default_head = head_type(self.num_classes, self.head_dim, loss=self.loss)
        self.default_head.init_weights()
        self.default_input = torch.ones((2, self.head_dim))
        self.default_ssl_input = {
            "labeled_weak": self.default_input,
            "labeled_strong": self.default_input,
            "unlabeled_weak": self.default_input,
            "unlabeled_strong": self.default_input,
        }
        self.default_gt = torch.zeros((2, self.num_classes))

    @e2e_pytest_unit
    def test_forward(self) -> None:
        result = self.default_head.forward_train(self.default_ssl_input, self.default_gt)
        assert "loss" in result
        assert "unlabeled_loss" in result
        assert result["loss"] >= 0

    @e2e_pytest_unit
    def test_simple_test(self) -> None:
        result = self.default_head.simple_test(self.default_input)
        assert result[0].shape[0] == self.num_classes

    @e2e_pytest_unit
    def test_angular(self, head_type) -> None:
        head = head_type(self.num_classes, self.head_dim, loss=self.loss, normalized=True)
        result = head.simple_test(self.default_input)
        assert result[0].shape[0] == self.num_classes

    @e2e_pytest_unit
    def test_neg_classes(self, head_type) -> None:
        with pytest.raises(ValueError):
            head_type(-1, self.head_dim, loss=self.loss, normalized=True)


class TestSemiNonLinearMultilabelClsHead(TestSemiLinearMultilabelClsHead):
    @pytest.fixture(autouse=True)
    def head_type(self):
        return SemiNonLinearMultilabelClsHead

    @e2e_pytest_unit
    def test_dropout(self, head_type) -> None:
        head = head_type(self.num_classes, self.head_dim, dropout=True)
        head.init_weights()
        assert len(head.classifier) > len(self.default_head.classifier)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

from dataclasses import dataclass

import pytest

from otx.core.ov.ops.builder import OperationRegistry
from otx.core.ov.ops.op import Attribute, Operation
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestOperationRegistry:
    @e2e_pytest_unit
    def test(self):
        registry = OperationRegistry("dummy", add_name_as_attr=True)
        OperationRegistry.REGISTERED_NAME_ATTR

        @dataclass
        class DummyAttributeV1(Attribute):
            pass

        class DummyV1(Operation[DummyAttributeV1]):
            TYPE = "dummy"
            VERSION = 1
            ATTRIBUTE_FACTORY = DummyAttributeV1

        registry.register()(DummyV1)
        assert getattr(DummyV1, OperationRegistry.REGISTERED_NAME_ATTR) == "DummyV1"

        with pytest.raises(KeyError):
            registry.register("another_dummy")(DummyV1)

        @dataclass
        class DummyAttributeV2(Attribute):
            pass

        class DummyV2(Operation[DummyAttributeV2]):
            TYPE = "dummy"
            VERSION = 2
            ATTRIBUTE_FACTORY = DummyAttributeV2

        registry.register()(DummyV2)
        assert getattr(DummyV2, OperationRegistry.REGISTERED_NAME_ATTR) == "DummyV2"

        assert DummyV1 == registry.get_by_name("DummyV1")
        assert DummyV1 == registry.get_by_type_version("dummy", 1)
        assert DummyV2 == registry.get_by_type_version("dummy", 2)

        with pytest.raises(KeyError):
            registry.get_by_type_version("dummy", 3)
        with pytest.raises(KeyError):
            registry.get_by_type_version("invalid", 1)



"""Functions for onnx adapters."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from functools import partial
from typing import Callable

import onnx
from onnx import ModelProto, NodeProto  # pylint: disable = no-name-in-module


def remove_nodes(model: ModelProto, predicate: Callable) -> ModelProto:
    """Remove nodes from ONNX model.

    Args:
        model (onnx.ModelProto): Input onnx model.
        predicate (Callable): A function to predicate a node.

    Returns:
        onnx.ModelProto: Modified onnx model.
    """
    # ! this doesn't handle inputs/outputs
    while True:
        connect = None
        for i, node in enumerate(model.graph.node):
            if predicate(node):
                assert len(node.input) == 1
                assert len(node.output) == 1
                connect = (node.input[0], node.output[0])
                del model.graph.node[i]
                break
        if not connect:
            break
        src, dst = connect
        for node in model.graph.node:
            for i, _input in enumerate(node.input):
                if _input == dst:
                    node.input[i] = src
    return model


def is_op(node: NodeProto, op_name) -> bool:
    """Check if an op is identity."""
    return node.op_type == op_name


def remove_node(model: ModelProto, op_name: str) -> ModelProto:  # noqa: C901
    """Remove identity node from an ONNX model.

    Args:
        model (onnx.ModelProto): Input onnx model.
        op_name (str): Operation name.
    """
    graph = model.graph

    def simplify_inputs():
        connect = None
        for _input in graph.input:
            for i, node in enumerate(graph.node):
                if node.op_type == op_name and node.input[0] == _input.name:
                    connect = (node.input[0], node.output[0])
                    del graph.node[i]
                    break
            if connect:
                break
        if not connect:
            return False
        src, dst = connect
        for node in graph.node:
            for i, input_name in enumerate(node.input):
                if input_name == dst:
                    node.input[i] = src
        # the input just changed won't be an output
        return True

    def simplify_outputs():
        connect = None
        for output in graph.output:
            for i, node in enumerate(graph.node):
                if node.op_type == op_name and node.output[0] == output.name:
                    connect = (node.input[0], node.output[0])
                    del graph.node[i]
                    break
            if connect:
                break
        if not connect:
            return False
        src, dst = connect
        for node in graph.node:
            for i, output_name in enumerate(node.output):
                if output_name == src:
                    node.output[i] = dst
            # the output just renamed may be someone's input
            for i, input_name in enumerate(node.input):
                if input_name == src:
                    node.input[i] = dst
        return True

    while simplify_inputs():
        pass

    while simplify_outputs():
        pass

    new_op = partial(is_op, op_name=op_name)
    remove_nodes(model, new_op)


def remove_nodes_by_op_type(onnx_model, op_type):
    """Remove all nodes of a specified op type from the ONNX model."""
    # TODO: support more nodes
    remove_node(onnx_model, op_type)
    onnx.checker.check_model(onnx_model)
    return onnx_model


def prepare_onnx_for_openvino(in_path, out_path):
    """Modify the specified ONNX model to be compatible with OpenVINO by removing 'Mark' op nodes."""
    onnx_model = onnx.load(in_path)
    onnx_model = remove_nodes_by_op_type(onnx_model, "Mark")
    onnx.checker.check_model(onnx_model)
    onnx.save(onnx_model, out_path)



"""HPO algorithm abstract class."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import tempfile
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Any, Dict, List, Optional, Union

from otx.hpo.search_space import SearchSpace
from otx.hpo.utils import check_mode_input, check_positive

logger = logging.getLogger(__name__)


class HpoBase(ABC):
    """Abstract class for HPO algorithms.

    This implements class for HPO algorithm base.
    Only common methods are implemented but not core algorithm of HPO.

    Args:
        search_space (Dict[str, Dict[str, Any]]): hyper parameter search space to find.
        save_path (Optional[str]): path where result of HPO is saved.
        mode (str, optinal): One of {min, max}. Determines whether objective is
                    minimizing or maximizing the metric attribute.
        num_trials (Optional[int]): How many training to conduct for HPO.
        num_workers (int): How many trains are executed in parallel.
        num_full_iterations (int): epoch for traninig after HPO.
        non_pure_train_ratio (float): ratio of validation time to (train time + validation time)
        full_dataset_size (int): train dataset size
        metric (str): Which score metric to use.
        expected_time_ratio (Optional[Union[int, float]]): Time to use for HPO.
                                            If HPO is configured automatically,
                                            HPO use time about exepected_time_ratio *
                                            train time after HPO times.
        maximum_resource (Optional[Union[int, float]]): Maximum resource to use for training each trial.
        subset_ratio (Optional[Union[float, int]]): ratio to how many train dataset to use for each trial.
                                     The lower value is, the faster the speed is.
                                     But If it's too low, HPO can be unstable.
        min_subset_size (int) : Minimum size of subset. Default value is 500.
        resume (bool): resume flag decide to use previous HPO results.
                       If HPO completed, you can just use optimized hyper parameters.
                       If HPO stopped in middle, you can resume in middle.
        prior_hyper_parameters (Optional[Union[Dict, List[Dict]]]) = Hyper parameters to try first.
        acceptable_additional_time_ratio (Union[float, int]) = Decide how much additional time can be acceptable.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        search_space: Dict[str, Dict[str, Any]],
        save_path: Optional[str] = None,
        mode: str = "max",
        num_trials: Optional[int] = None,
        num_workers: int = 1,
        num_full_iterations: Union[int, float] = 1,
        non_pure_train_ratio: float = 0.2,
        full_dataset_size: int = 0,
        metric: str = "mAP",
        expected_time_ratio: Optional[Union[int, float]] = None,
        maximum_resource: Optional[Union[int, float]] = None,
        subset_ratio: Optional[Union[float, int]] = None,
        min_subset_size: int = 500,
        resume: bool = False,
        prior_hyper_parameters: Optional[Union[Dict, List[Dict]]] = None,
        acceptable_additional_time_ratio: Union[float, int] = 1.0,
    ):
        # pylint: disable=too-many-arguments, too-many-locals
        check_mode_input(mode)
        check_positive(full_dataset_size, "full_dataset_size")
        check_positive(num_full_iterations, "num_full_iterations")
        if not 0 < non_pure_train_ratio <= 1:
            raise ValueError(
                "non_pure_train_ratio should be greater than 0 and lesser than or equal to 1."
                f" Your value is {subset_ratio}"
            )
        if maximum_resource is not None:
            check_positive(maximum_resource, "maximum_resource")
        if num_trials is not None:
            check_positive(num_trials, "num_trials")
        check_positive(num_workers, "num_workers")
        if subset_ratio is not None:
            if not 0 < subset_ratio <= 1:
                raise ValueError(
                    "subset_ratio should be greater than 0 and lesser than or equal to 1."
                    f" Your value is {subset_ratio}"
                )

        if save_path is None:
            save_path = tempfile.mkdtemp(prefix="OTX-hpo-")
        self.save_path = save_path
        self.search_space = SearchSpace(search_space)
        self.mode = mode
        self.num_trials = num_trials
        self.num_workers = num_workers
        self.num_full_iterations = num_full_iterations
        self.non_pure_train_ratio = non_pure_train_ratio
        self.full_dataset_size = full_dataset_size
        self.expected_time_ratio = expected_time_ratio
        self.maximum_resource: Optional[Union[int, float]] = maximum_resource
        self.subset_ratio = subset_ratio
        self.min_subset_size = min_subset_size
        self.resume = resume
        self.hpo_status: dict = {}
        self.metric = metric
        self.acceptable_additional_time_ratio = acceptable_additional_time_ratio
        if prior_hyper_parameters is None:
            prior_hyper_parameters = []
        elif isinstance(prior_hyper_parameters, dict):
            prior_hyper_parameters = [prior_hyper_parameters]
        self.prior_hyper_parameters = prior_hyper_parameters

    @abstractmethod
    def print_result(self):
        """Print a HPO algorithm result."""
        raise NotImplementedError

    @abstractmethod
    def save_results(self):
        """Save a HPO algorithm result."""
        raise NotImplementedError

    @abstractmethod
    def is_done(self):
        """Check whether HPO algorithm is done."""
        raise NotImplementedError

    @abstractmethod
    def get_next_sample(self):
        """Get next sample to train."""
        raise NotImplementedError

    @abstractmethod
    def auto_config(self):
        """Configure HPO algorithm automatically."""
        raise NotImplementedError

    @abstractmethod
    def get_progress(self):
        """Get current progress of HPO algorithm."""
        raise NotImplementedError

    @abstractmethod
    def report_score(self, score, resource, trial_id, done):
        """Report a score to HPO algorithm."""
        raise NotImplementedError

    @abstractmethod
    def get_best_config(self):
        """Get best config of HPO algorithm."""
        raise NotImplementedError


class Trial:
    """Trial to train with given hyper parameters.

    Args:
        trial_id (Any): Trial id.
        configuration (Dict): Configuration to train with.
        train_environment (Optional[Dict], optional): Train environment for the trial. Defaults to None.
    """

    def __init__(self, trial_id: Any, configuration: Dict, train_environment: Optional[Dict] = None):
        self._id = trial_id
        self._configuration = configuration
        self.score: Dict[Union[float, int], Union[float, int]] = {}
        self._train_environment = train_environment
        self._iteration = None
        self.status: TrialStatus = TrialStatus.READY

    @property
    def id(self):
        """Trial id."""
        return self._id

    @property
    def configuration(self):
        """Configuration to train with."""
        return self._configuration

    @property
    def iteration(self):
        """Iteration to use for training."""
        return self._iteration

    @iteration.setter
    def iteration(self, val):
        """Setter for iteration."""
        check_positive(val, "iteration")
        self._iteration = val

    @property
    def train_environment(self):
        """Train environment for the trial."""
        return self._train_environment

    def get_train_configuration(self) -> Dict[str, Any]:
        """Get configurations needed to trian."""
        self._configuration["iterations"] = self.iteration
        return {"id": self.id, "configuration": self.configuration, "train_environment": self.train_environment}

    def register_score(self, score: Union[int, float], resource: Union[int, float]):
        """Register score to the trial.

        Args:
            score (Union[int, float]): Score to register.
            resource (Union[int, float]): Resource used to get score. It should be positive.
        """
        check_positive(resource, "resource")
        self.score[resource] = score

    def get_best_score(
        self, mode: str = "max", resource_limit: Optional[Union[float, int]] = None
    ) -> Optional[Union[float, int]]:
        """Get best score of the trial.

        Args:
            mode (str, optional): Decide which is better between highest score or lowest score. Defaults to "max".
            resource_limit (Optional[Union[float, int]], optional): Find a best score among the score at resource
                                                                    lower than this value. Defaults to None.

        Returns:
            Optional[Union[float, int]]: Best score. If there is no score, return None.
        """
        check_mode_input(mode)

        if resource_limit is None:
            scores = self.score.values()
        else:
            scores = [val for key, val in self.score.items() if key <= resource_limit]  # type: ignore

        if len(scores) == 0:
            return None

        if mode == "max":
            return max(scores)
        return min(scores)

    def get_progress(self) -> Union[float, int]:
        """Get a progress of the trial.

        Returns:
            Union[float, int]: How many resource is used for the trial.
        """
        if len(self.score) == 0:
            return 0
        return max(self.score.keys())

    def save_results(self, save_path: str):
        """Save a result in the 'save_path'.

        Args:
            save_path (str): Path where to save a result.
        """
        results = {
            "id": self.id,
            "configuration": self.configuration,
            "train_environment": self.train_environment,
            "score": self.score,
        }

        with open(save_path, "w", encoding="utf-8") as f:
            json.dump(results, f)

    def finalize(self):
        """Let the trial know that training is done.

        If the trial isn't trained until given resource, then make it pretend to train until resouce.
        """
        if self.get_progress() < self.iteration:
            best_score = self.get_best_score()
            if best_score is None:
                raise RuntimeError(f"Although {self.id} trial doesn't report any score but it's done")
            self.register_score(best_score, self.iteration)

    def is_done(self):
        """Check the trial is done."""
        if self.iteration is None:
            raise ValueError("iteration isn't set yet.")
        return self.get_progress() >= self.iteration


class TrialStatus(IntEnum):
    """Enum class for trial status."""

    UNKNOWN = -1
    READY = 0
    RUNNING = 1
    STOP = 2
    CUDAOOM = 3



import copy
import math

import pytest

from otx.hpo.search_space import SearchSpace, SingleSearchSpace
from tests.test_suite.e2e_test_system import e2e_pytest_component

ALL_TYPE = ["uniform", "loguniform", "quniform", "qloguniform", "choice"]
NOT_CATEGORICAL_TYPE = ["uniform", "loguniform", "quniform", "qloguniform"]
USE_LOG_SCALE_TYPE = ["loguniform", "qloguniform"]
USE_QUANTIZED_STEP_TYPE = ["quniform", "qloguniform"]


def make_single_search_space_uniform_good_arguments():
    return [
        {"type": "uniform", "min": 1, "max": 100},
        {"type": "uniform", "min": 1.12, "max": 53.221},
        {"type": "uniform", "min": -300, "max": -299},
        {"type": "uniform", "min": -12.543, "max": 15.233},
    ]


def add_step_from_args(args, type_name=None):
    new_args = []
    for arg in args:
        step = (arg["max"] - arg["min"]) / 2
        arg["step"] = step
        if type_name is not None:
            arg["type"] = type_name
        new_args.append(arg)

    return new_args


def add_log_base_from_args(args, type_name):
    new_args = []
    for log_base in [2, 10]:
        for arg in args:
            if arg["min"] <= 0:
                continue
            arg["log_base"] = log_base
            if type_name is not None:
                arg["type"] = type_name
            new_args.append(arg)

    return new_args


def make_single_search_space_quniform_good_arguments():
    return add_step_from_args(make_single_search_space_uniform_good_arguments(), "quniform")


def make_single_search_space_loguniform_good_arguments():
    return add_log_base_from_args(make_single_search_space_uniform_good_arguments(), "loguniform")


def make_single_search_space_qloguniform_good_arguments():
    return add_log_base_from_args(make_single_search_space_quniform_good_arguments(), "qloguniform")


def make_single_search_space_choice_good_arguments():
    return [{"type": "choice", "choice_list": ("abc", "def")}, {"type": "choice", "choice_list": [1, 2, 3]}]


MAKE_GOOD_ARGS = {
    "uniform": make_single_search_space_uniform_good_arguments,
    "quniform": make_single_search_space_quniform_good_arguments,
    "loguniform": make_single_search_space_loguniform_good_arguments,
    "qloguniform": make_single_search_space_qloguniform_good_arguments,
    "choice": make_single_search_space_choice_good_arguments,
}


def get_wrong_arg(original_arg, attr_names, values, errors):
    wrong_arg = copy.deepcopy(original_arg)
    if not isinstance(attr_names, (list, tuple)):
        attr_names = [attr_names]
        values = [values]

    for attr_name, value in zip(attr_names, values):
        wrong_arg[attr_name] = value
    wrong_arg["error"] = errors

    return wrong_arg


def make_arg_minmax_wrong(arg):
    args = []
    # value is None
    for attr_name in ["min", "max"]:
        args.append(get_wrong_arg(arg, attr_name, None, (ValueError, TypeError)))

    # min is greater or same than max
    for val in [(3, 1), (5.5, 2.3), (1, 1), (2.0, 2.0)]:
        args.append(get_wrong_arg(arg, ["min", "max"], val, ValueError))

    # value is minus although using log scale
    if "log_base" in arg:
        for val in [(-20, -12), (-12.124, 10), (0, 3)]:
            args.append(get_wrong_arg(arg, ["min", "max"], val, ValueError))

    return args


def make_arg_step_wrong(arg):
    args = []

    # vlaue is None
    args.append(get_wrong_arg(arg, "step", None, (TypeError, ValueError)))

    # step is too big
    args.append(get_wrong_arg(arg, "step", arg["max"] - arg["min"] + 1, (TypeError, ValueError)))

    return args


def make_arg_logbase_wrong(arg):
    args = []

    # too small value
    for val in [1, 0, -1]:
        args.append(get_wrong_arg(arg, "log_base", val, ValueError))

    return args


def make_arg_choicelist_wrong(arg):
    args = []

    # vlaue is None
    args.append(get_wrong_arg(arg, "choice_list", None, (TypeError, ValueError)))
    # few elements

    for val in [[], [1]]:
        args.append(get_wrong_arg(arg, "choice_list", val, ValueError))

    return args


def make_arg_type_wrong(arg):
    args = []

    # wrong type
    for val in ["wrong_type", 12, 1.24, [1, 2]]:
        args.append(get_wrong_arg(arg, "type", val, ValueError))

    return args


class TestSingleSearchSpace:
    @e2e_pytest_component
    @pytest.mark.parametrize("type", ALL_TYPE)
    def test_init_with_good_input(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            SingleSearchSpace(**arg)

    @e2e_pytest_component
    @pytest.mark.parametrize("type", NOT_CATEGORICAL_TYPE)
    def test_init_wrong_minmax_arg(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            wrong_args = make_arg_minmax_wrong(arg)
            for wrong_arg in wrong_args:
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                with pytest.raises(errors):
                    SingleSearchSpace(**wrong_arg)

    @e2e_pytest_component
    @pytest.mark.parametrize("type", USE_QUANTIZED_STEP_TYPE)
    def test_init_wrong_step_arg(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            wrong_args = make_arg_step_wrong(arg)
            for wrong_arg in wrong_args:
                print(wrong_arg)
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                with pytest.raises(errors):
                    SingleSearchSpace(**wrong_arg)

    @e2e_pytest_component
    @pytest.mark.parametrize("type", USE_LOG_SCALE_TYPE)
    def test_init_wrong_log_base_arg(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            wrong_args = make_arg_logbase_wrong(arg)
            for wrong_arg in wrong_args:
                print(wrong_arg)
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                with pytest.raises(errors):
                    SingleSearchSpace(**wrong_arg)

    @e2e_pytest_component
    def test_init_wrong_choice_list_arg(self):
        args = MAKE_GOOD_ARGS["choice"]()
        for arg in args:
            wrong_args = make_arg_choicelist_wrong(arg)
            for wrong_arg in wrong_args:
                print(wrong_arg)
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                with pytest.raises(errors):
                    SingleSearchSpace(**wrong_arg)

    @e2e_pytest_component
    def test_init_with_wrong_type(self):
        for type in ALL_TYPE:
            args = MAKE_GOOD_ARGS[type]()
            for arg in args:
                wrong_args = make_arg_type_wrong(arg)
                for wrong_arg in wrong_args:
                    errors = wrong_arg["error"]
                    del wrong_arg["error"]
                    with pytest.raises(errors):
                        SingleSearchSpace(**wrong_arg)

    @e2e_pytest_component
    def test_set_value_normally(self):
        args = []
        for type in ALL_TYPE:
            args.extend(MAKE_GOOD_ARGS[type]())
        cur_arg = args.pop(0)
        for new_arg in args:
            sss = SingleSearchSpace(**cur_arg)
            sss.set_value(**new_arg)
            cur_arg = new_arg

    @e2e_pytest_component
    @pytest.mark.parametrize("type", NOT_CATEGORICAL_TYPE)
    def test_set_value_wrong_minmax_arg(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            wrong_args = make_arg_minmax_wrong(arg)
            for wrong_arg in wrong_args:
                if wrong_arg["min"] is None or wrong_arg["max"] is None:
                    continue
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                print(wrong_arg)
                with pytest.raises(errors):
                    sss.set_value(**wrong_arg)

    @e2e_pytest_component
    @pytest.mark.parametrize("type", USE_QUANTIZED_STEP_TYPE)
    def test_set_value_wrong_step_arg(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            wrong_args = make_arg_step_wrong(arg)
            for wrong_arg in wrong_args:
                if wrong_arg["step"] is None:
                    continue
                print(wrong_arg)
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                with pytest.raises(errors):
                    sss.set_value(**wrong_arg)

    @e2e_pytest_component
    @pytest.mark.parametrize("type", USE_LOG_SCALE_TYPE)
    def test_set_value_wrong_log_base_arg(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            wrong_args = make_arg_logbase_wrong(arg)
            for wrong_arg in wrong_args:
                if wrong_arg["log_base"] is None:
                    continue
                print(wrong_arg)
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                with pytest.raises(errors):
                    sss.set_value(**wrong_arg)

    @e2e_pytest_component
    def test_set_value_wrong_choice_list_arg(self):
        args = MAKE_GOOD_ARGS["choice"]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            wrong_args = make_arg_choicelist_wrong(arg)
            for wrong_arg in wrong_args:
                if wrong_arg["choice_list"] is None:
                    continue
                print(wrong_arg)
                errors = wrong_arg["error"]
                del wrong_arg["error"]
                with pytest.raises(errors):
                    sss.set_value(**wrong_arg)

    @e2e_pytest_component
    def test_set_value_with_wrong_type(self):
        for type in ALL_TYPE:
            args = MAKE_GOOD_ARGS[type]()
            for arg in args:
                sss = SingleSearchSpace(**arg)
                wrong_args = make_arg_type_wrong(arg)
                for wrong_arg in wrong_args:
                    errors = wrong_arg["error"]
                    del wrong_arg["error"]
                    with pytest.raises(errors):
                        sss.set_value(**wrong_arg)

    @e2e_pytest_component
    def test_align_min_max_to_choice_list_if_categorical(self):
        args = MAKE_GOOD_ARGS["choice"]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            sss._min = -100
            sss._max = -100
            sss._align_min_max_to_choice_list_if_categorical()
            assert sss.min == 0
            assert sss.max == len(arg["choice_list"]) - 1

    @e2e_pytest_component
    @pytest.mark.parametrize("type", ALL_TYPE)
    def test_is_categorical(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            if type in NOT_CATEGORICAL_TYPE:
                assert not sss.is_categorical()
            else:
                assert sss.is_categorical()

    @e2e_pytest_component
    @pytest.mark.parametrize("type", ALL_TYPE)
    def test_use_quantized_step(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            if type in USE_QUANTIZED_STEP_TYPE:
                assert sss.use_quantized_step()
            else:
                assert not sss.use_quantized_step()

    @e2e_pytest_component
    @pytest.mark.parametrize("type", ALL_TYPE)
    def test_use_log_scale(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            if type in USE_LOG_SCALE_TYPE:
                assert sss.use_log_scale()
            else:
                assert not sss.use_log_scale()

    @e2e_pytest_component
    @pytest.mark.parametrize("type", ALL_TYPE)
    def test_lower_space_upper_space(self, type):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            if type == "choice":
                assert sss.lower_space() == 0
                assert sss.upper_space() == len(sss.choice_list) - 1
            else:
                min = arg["min"]
                max = arg["max"]
                if type in USE_LOG_SCALE_TYPE:
                    log_base = arg["log_base"]
                    assert sss.lower_space() == math.log(min, log_base)
                    assert sss.upper_space() == math.log(max, log_base)
                else:
                    assert sss.lower_space() == min
                    assert sss.upper_space() == max

    @e2e_pytest_component
    @pytest.mark.parametrize("type", ALL_TYPE)
    @pytest.mark.parametrize("number", [2.3, 15])
    def test_space_to_real(self, type, number):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            if type == "choice":
                choice_list = arg["choice_list"]
                ret = sss.space_to_real(number)
                expected_ret = min(max(int(number), 0), len(choice_list) - 1)
                assert ret == sss.choice_list[expected_ret]
            else:
                ret = sss.space_to_real(number)
                expected_ret = number

                if type in USE_LOG_SCALE_TYPE:
                    log_base = arg["log_base"]
                    expected_ret = log_base**expected_ret
                if type in USE_QUANTIZED_STEP_TYPE:
                    step = arg["step"]
                    gap = sss.min % step
                    expected_ret = round((expected_ret - gap) / step) * step + gap
                assert ret == expected_ret

    @e2e_pytest_component
    @pytest.mark.parametrize("type", ALL_TYPE)
    @pytest.mark.parametrize("number", [10, 512.3])
    def test_real_to_space(self, type, number):
        args = MAKE_GOOD_ARGS[type]()
        for arg in args:
            sss = SingleSearchSpace(**arg)
            if type in USE_LOG_SCALE_TYPE:
                log_base = arg["log_base"]
                assert sss.real_to_space(number) == math.log(number, log_base)
            else:
                assert sss.real_to_space(number) == number


class TestSearchSpace:
    @staticmethod
    def get_search_space_depending_on_type(types, range_format=False):
        if not isinstance(types, (list, tuple)):
            types = [types]
        search_space = {}

        if "uniform" in types:
            TestSearchSpace.add_uniform_search_space(search_space, range_format)
        if "quniform" in types:
            TestSearchSpace.add_quniform_search_space(search_space, range_format)
        if "loguniform" in types:
            TestSearchSpace.add_loguniform_search_space(search_space, range_format)
        if "qloguniform" in types:
            TestSearchSpace.add_qloguniform_search_space(search_space, range_format)
        if "choice" in types:
            TestSearchSpace.add_choice_search_space(search_space)

        return search_space

    @staticmethod
    def add_uniform_search_space(search_space, range_format=False):
        search_space["uniform_search_space"] = {"param_type": "uniform"}
        if range_format:
            search_space["uniform_search_space"]["range"] = [1, 10]
        else:
            search_space["uniform_search_space"].update({"min": 1, "max": 10})

    @staticmethod
    def add_quniform_search_space(search_space, range_format=False):
        search_space["quniform_search_space"] = {"param_type": "quniform"}
        if range_format:
            search_space["quniform_search_space"]["range"] = [1, 10, 3]
        else:
            search_space["quniform_search_space"].update({"min": 1, "max": 10, "step": 3})

    @staticmethod
    def add_loguniform_search_space(search_space, range_format=False):
        search_space["loguniform_search_space"] = {"param_type": "loguniform"}
        if range_format:
            search_space["loguniform_search_space"]["range"] = [1, 10, 2]
        else:
            search_space["loguniform_search_space"].update({"min": 1, "max": 10, "log_base": 2})

    @staticmethod
    def add_qloguniform_search_space(search_space, range_format=False):
        search_space["qloguniform_search_space"] = {"param_type": "qloguniform"}
        if range_format:
            search_space["qloguniform_search_space"]["range"] = [1, 10, 3, 2]
        else:
            search_space["qloguniform_search_space"].update({"min": 1, "max": 10, "step": 3, "log_base": 2})

    @staticmethod
    def add_choice_search_space(search_space):
        search_space["choice_search_space"] = {
            "param_type": "choice",
            "choice_list": ["somevalue1", "somevalue2", "somevalue3"],
        }

    @pytest.fixture
    def search_space_with_all_types(self):
        return SearchSpace(self.get_search_space_depending_on_type(ALL_TYPE))

    @e2e_pytest_component
    def test_init_with_range_format_argument(self):
        search_space = self.get_search_space_depending_on_type(ALL_TYPE, True)
        ss = SearchSpace(search_space)
        assert ss is not None

    @e2e_pytest_component
    @pytest.mark.parametrize("type", NOT_CATEGORICAL_TYPE)
    def test_init_with_insufficient_range_arguments(self, type):
        search_space = self.get_search_space_depending_on_type(type, True)
        if type in USE_LOG_SCALE_TYPE:
            num_to_delete = 2
        else:
            num_to_delete = 1
        search_space[f"{type}_search_space"]["range"] = search_space[f"{type}_search_space"]["range"][:-num_to_delete]
        with pytest.raises(ValueError):
            SearchSpace(search_space)

    @e2e_pytest_component
    def test_init_both_format_exists(self):
        search_space = self.get_search_space_depending_on_type(ALL_TYPE)
        range_format = self.get_search_space_depending_on_type(ALL_TYPE, True)
        for key, val in search_space.items():
            val.update(range_format[key])
        SearchSpace(search_space)

    @e2e_pytest_component
    def test_get_item_available(self, search_space_with_all_types):
        for type in ALL_TYPE:
            search_space_with_all_types[f"{type}_search_space"]

    @e2e_pytest_component
    def test_iteratble(self, search_space_with_all_types):
        for val in search_space_with_all_types:
            pass

    @e2e_pytest_component
    def test_len_is_available(self, search_space_with_all_types):
        assert len(search_space_with_all_types) == 5

    @e2e_pytest_component
    @pytest.mark.parametrize("choice_exist", [True, False])
    def test_has_categorical_param(self, search_space_with_all_types, choice_exist):
        if choice_exist:
            search_space = self.get_search_space_depending_on_type(ALL_TYPE)
        else:
            search_space = self.get_search_space_depending_on_type(NOT_CATEGORICAL_TYPE)
        ss = SearchSpace(search_space)

        assert ss.has_categorical_param() == choice_exist

    @e2e_pytest_component
    def test_get_real_config_with_proper_argument(self, search_space_with_all_types):
        # search space configuration
        step = 3
        log_base = 2
        min_val = 1
        requested_val = 3.2

        config = {f"{type}_search_space": requested_val for type in ALL_TYPE}
        real_space = search_space_with_all_types.get_real_config(config)

        for key, val in real_space.items():
            rescaled_requested_val = requested_val
            if key in ["loguniform_search_space", "qloguniform_search_space"]:
                rescaled_requested_val = log_base**requested_val
            if key in ["quniform_search_space", "qloguniform_search_space"]:
                gap = min_val % step
                rescaled_requested_val = round((rescaled_requested_val - gap) / step) * step + gap
            if key == "choice_search_space":
                choice_list = search_space_with_all_types[key].choice_list
                idx = max(min(int(rescaled_requested_val), len(choice_list) - 1), 0)
                rescaled_requested_val = choice_list[idx]

            assert val == rescaled_requested_val

    @e2e_pytest_component
    @pytest.mark.parametrize("wrong_name", ["wrong_name", 1, 3.2])
    def test_get_real_config_with_wrong_name_config(self, search_space_with_all_types, wrong_name):
        config = {wrong_name: 3.2}
        with pytest.raises(KeyError):
            search_space_with_all_types.get_real_config(config)

    @e2e_pytest_component
    def test_get_space_config_with_proper_argument(self, search_space_with_all_types):
        # search space configuration
        log_base = 2
        requested_val = 10

        config = {f"{type}_search_space": requested_val for type in ALL_TYPE}
        real_space = search_space_with_all_types.get_space_config(config)

        for key, val in real_space.items():
            rescaled_requested_val = requested_val
            if key in ["loguniform_search_space", "qloguniform_search_space"]:
                rescaled_requested_val = math.log(requested_val, log_base)

            assert val == rescaled_requested_val

    @e2e_pytest_component
    @pytest.mark.parametrize("wrong_name", ["wrong_name", 1, 3.2])
    def test_get_space_config_with_wrong_name_config(self, search_space_with_all_types, wrong_name):
        config = {wrong_name: 3.2}
        with pytest.raises(KeyError):
            search_space_with_all_types.get_space_config(config)

    @e2e_pytest_component
    def test_get_bayeopt_search_space(self, search_space_with_all_types):
        bayes_opt_format = search_space_with_all_types.get_bayeopt_search_space()

        for val in bayes_opt_format.values():
            assert len(val) == 2
            min_val, max_val = val
            assert min_val < max_val

    @e2e_pytest_component
    def test_convert_from_zero_one_scale_to_real_space_with_good_args(self, search_space_with_all_types):
        config = {}
        for key in search_space_with_all_types:
            config[key] = 0.5
        search_space_with_all_types.convert_from_zero_one_scale_to_real_space(config)

    @e2e_pytest_component
    @pytest.mark.parametrize("config", ["wrong_value", [1, 3, 4], (1, 2)])
    def test_convert_from_zero_one_scale_to_real_space_with_bad_arg_type(self, search_space_with_all_types, config):
        with pytest.raises(AttributeError):
            search_space_with_all_types.convert_from_zero_one_scale_to_real_space(config)



"""Registry Class for otx.core.ov."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict, Optional


class Registry:
    """Registry Class for OMZ model."""

    REGISTERED_NAME_ATTR = "_registered_name"

    def __init__(self, name, add_name_as_attr=False):
        self._name = name
        self._registry_dict = {}
        self._add_name_as_attr = add_name_as_attr

    @property
    def registry_dict(self) -> Dict[Any, Any]:
        """Dictionary of registered module."""
        return self._registry_dict

    def _register(self, obj: Any, name: Any):
        """Register obj with name."""
        if name in self._registry_dict:
            raise KeyError(f"{name} is already registered in {self._name}")
        self._registry_dict[name] = obj

    def register(self, name: Optional[Any] = None):
        """Register from name."""

        def wrap(obj):
            cls_name = name
            if cls_name is None:
                cls_name = obj.__name__
            if self._add_name_as_attr:
                setattr(obj, self.REGISTERED_NAME_ATTR, cls_name)
            self._register(obj, cls_name)
            return obj

        return wrap

    def get(self, key: Any) -> Any:
        """Get from module name (key)."""
        if key not in self._registry_dict:
            self._key_not_found(key)
        return self._registry_dict[key]

    def _key_not_found(self, key: Any):
        """Raise KeyError when key not founded."""
        raise KeyError(f"{key} is not found in {self._name}")

    def __contains__(self, item):
        """Check containing of item."""
        return item in self._registry_dict.values()



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
from attr import _make, validators

from otx.api.configuration.elements.configurable_enum import ConfigurableEnum
from otx.api.configuration.elements.primitive_parameters import (
    boolean_attribute,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    float_selectable,
    selectable,
    set_common_metadata,
    string_attribute,
)
from otx.api.configuration.enums import AutoHPOState, ConfigElementType, ModelLifecycle
from otx.api.configuration.ui_rules import NullUIRules, Rule, UIRules
from tests.unit.api.configuration.dummy_config import SomeEnumSelectable
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestPrimitiveParameters:
    ui_rules = UIRules(rules=[Rule(parameter="rule parameter", value=1)])

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_set_common_metadata(self):
        """
        <b>Description:</b>
        Check "set_common_metadata" function

        <b>Input data:</b>
        "default_value" int, float, str, bool or ConfigurableEnum element, "header" string, "description" string,
        "warning" string, "editable" bool value, "affects_outcome_of" ModelLifecycle element, "ui_rules" UIRules object,
        "visible_in_ui" bool value, "parameter_type" ConfigElementType element

        <b>Expected results:</b>
        Test passes if dictionary returned by "set_common_metadata" function is equal to expected
        """
        header = "test header"
        description = "test description"
        warning = "test warning"
        editable = False
        affects_outcome_of = ModelLifecycle.TESTING
        ui_rules = self.ui_rules
        visible_in_ui = True
        parameter_type = ConfigElementType.CONFIGURABLE_PARAMETERS
        auto_hpo_state = AutoHPOState.POSSIBLE

        for default_value in [
            5,  # int "default_value"
            1.3,  # float "default_value"
            "default value string",  # str "default_value"
            False,  # bool "default_value"
            SomeEnumSelectable.TEST_2,  # ConfigurableEnum "default_value"
        ]:
            assert set_common_metadata(
                default_value=default_value,
                header=header,
                description=description,
                warning=warning,
                editable=editable,
                affects_outcome_of=affects_outcome_of,
                ui_rules=ui_rules,
                visible_in_ui=visible_in_ui,
                parameter_type=parameter_type,
                auto_hpo_state=auto_hpo_state,
                auto_hpo_value=default_value,
            ) == {
                "default_value": default_value,
                "description": description,
                "header": header,
                "warning": warning,
                "editable": editable,
                "visible_in_ui": visible_in_ui,
                "affects_outcome_of": affects_outcome_of,
                "ui_rules": ui_rules,
                "type": parameter_type,
                "auto_hpo_state": auto_hpo_state,
                "auto_hpo_value": default_value,
            }

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_configurable_integer(self):
        """
        <b>Description:</b>
        Check "configurable_integer" function

        <b>Input data:</b>
        "default_value" int, "header" string, "min_value" int, "max_value": int, "description" string, "warning" string,
        "editable" bool, "visible_in_ui" bool, "affects_outcome_of" ModelLifecycle element, "ui_rules" UIRules object

        <b>Expected results:</b>
        Test passes if _CountingAttr object returned by "configurable_integer" function is equal to expected

        <b>Steps</b>
        1. Check _CountingAttr object returned by "configurable_integer" function for default values of optional
        parameters
        2. Check _CountingAttr object returned by "configurable_integer" function for specified values of optional
        parameters
        """

        def check_configurable_integer(
            integer_instance,
            expected_min_value: int = 0,
            expected_max_value: int = 255,
            expected_description: str = "Default integer description",
            expected_warning: str = None,
            expected_editable: bool = True,
            expected_visible_in_ui: bool = True,
            expected_affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
            expected_ui_rules: UIRules = NullUIRules(),
            expected_auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
            expected_auto_hpo_value: int = None,
        ):
            expected_metadata = {
                "default_value": 100,
                "description": expected_description,
                "header": "configurable_integer header",
                "warning": expected_warning,
                "editable": expected_editable,
                "visible_in_ui": expected_visible_in_ui,
                "affects_outcome_of": expected_affects_outcome_of,
                "ui_rules": expected_ui_rules,
                "type": ConfigElementType.INTEGER,
                "min_value": expected_min_value,
                "max_value": expected_max_value,
                "auto_hpo_state": expected_auto_hpo_state,
                "auto_hpo_value": expected_auto_hpo_value,
            }
            assert isinstance(integer_instance, _make._CountingAttr)
            assert integer_instance._default == 100
            assert integer_instance.type == int
            assert len(integer_instance._validator._validators) == 2
            assert integer_instance.metadata == expected_metadata

        # Checking _CountingAttr object returned by "configurable_integer" for default values of optional parameters
        default_value = 100
        header = "configurable_integer header"
        actual_integer = configurable_integer(default_value=default_value, header=header)
        check_configurable_integer(integer_instance=actual_integer)  # type: ignore
        # Checking _CountingAttr object returned by "configurable_integer" for specified values of optional parameters
        min_value = 10
        max_value = 200
        description = "configurable_integer description"
        warning = "configurable_integer warning"
        editable = False
        visible_in_ui = False
        affects_outcome_of = ModelLifecycle.TESTING
        ui_rules = self.ui_rules
        auto_hpo_state = AutoHPOState.POSSIBLE
        auto_hpo_value = min_value

        actual_integer = configurable_integer(
            default_value=default_value,
            header=header,
            min_value=min_value,
            max_value=max_value,
            description=description,
            warning=warning,
            editable=editable,
            visible_in_ui=visible_in_ui,
            affects_outcome_of=affects_outcome_of,
            ui_rules=ui_rules,
            auto_hpo_value=auto_hpo_value,
            auto_hpo_state=auto_hpo_state,
        )
        check_configurable_integer(
            integer_instance=actual_integer,  # type: ignore
            expected_min_value=min_value,
            expected_max_value=max_value,
            expected_description=description,
            expected_warning=warning,
            expected_editable=editable,
            expected_visible_in_ui=visible_in_ui,
            expected_affects_outcome_of=affects_outcome_of,
            expected_ui_rules=ui_rules,
            expected_auto_hpo_state=auto_hpo_state,
            expected_auto_hpo_value=auto_hpo_value,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_configurable_float(self):
        """
        <b>Description:</b>
        Check "configurable_float" function

        <b>Input data:</b>
        "default_value" float, "header" string, "min_value" float, "max_value": float, "description" string,
        "warning" string, "editable" bool, "visible_in_ui" bool, "affects_outcome_of" ModelLifecycle element,
        "ui_rules" UIRules object

        <b>Expected results:</b>
        Test passes if _CountingAttr object returned by "configurable_float" function is equal to expected

        <b>Steps</b>
        1. Check _CountingAttr object returned by "configurable_float" function for default values of optional
        parameters
        2. Check _CountingAttr object returned by "configurable_float" function for specified values of optional
        parameters
        """

        def check_configurable_float(
            float_instance,
            expected_min_value: float = 0.0,
            expected_max_value: float = 255.0,
            expected_description: str = "Default float description",
            expected_warning: str = None,
            expected_editable: bool = True,
            expected_visible_in_ui: bool = True,
            expected_affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
            expected_ui_rules: UIRules = NullUIRules(),
            expected_auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
            expected_auto_hpo_value: float = None,
        ):
            expected_metadata = {
                "default_value": 100.1,
                "description": expected_description,
                "header": "configurable_float header",
                "warning": expected_warning,
                "editable": expected_editable,
                "visible_in_ui": expected_visible_in_ui,
                "affects_outcome_of": expected_affects_outcome_of,
                "ui_rules": expected_ui_rules,
                "type": ConfigElementType.FLOAT,
                "min_value": expected_min_value,
                "max_value": expected_max_value,
                "auto_hpo_state": expected_auto_hpo_state,
                "auto_hpo_value": expected_auto_hpo_value,
            }
            assert isinstance(float_instance, _make._CountingAttr)
            assert float_instance._default == 100.1
            assert float_instance.type == float
            assert float_instance.metadata == expected_metadata

        # Checking _CountingAttr object returned by "configurable_float" for default values of optional parameters
        default_value = 100.1
        header = "configurable_float header"
        actual_float = configurable_float(default_value=default_value, header=header)
        check_configurable_float(float_instance=actual_float)  # type: ignore
        # Checking _CountingAttr object returned by "configurable_float" for specified values of optional parameters
        min_value = 0.1
        max_value = 160.2
        description = "configurable_float description"
        warning = "configurable_float warning"
        editable = False
        visible_in_ui = False
        affects_outcome_of = ModelLifecycle.TESTING
        ui_rules = self.ui_rules
        auto_hpo_state = AutoHPOState.POSSIBLE
        auto_hpo_value = min_value

        actual_float = configurable_float(
            default_value=default_value,
            header=header,
            min_value=min_value,
            max_value=max_value,
            description=description,
            warning=warning,
            editable=editable,
            visible_in_ui=visible_in_ui,
            affects_outcome_of=affects_outcome_of,
            ui_rules=ui_rules,
            auto_hpo_value=auto_hpo_value,
            auto_hpo_state=auto_hpo_state,
        )
        check_configurable_float(
            float_instance=actual_float,  # type: ignore
            expected_min_value=min_value,
            expected_max_value=max_value,
            expected_description=description,
            expected_warning=warning,
            expected_editable=editable,
            expected_visible_in_ui=visible_in_ui,
            expected_affects_outcome_of=affects_outcome_of,
            expected_ui_rules=ui_rules,
            expected_auto_hpo_state=auto_hpo_state,
            expected_auto_hpo_value=auto_hpo_value,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_configurable_boolean(self):
        """
        <b>Description:</b>
        Check "configurable_boolean" function

        <b>Input data:</b>
        "default_value" bool, "header" string, "description" string, "warning" string, "editable" bool,
        "visible_in_ui" bool, "affects_outcome_of" ModelLifecycle element, "ui_rules" UIRules object

        <b>Expected results:</b>
        Test passes if _CountingAttr object returned by "configurable_boolean" function is equal to expected

        <b>Steps</b>
        1. Check _CountingAttr object returned by "configurable_boolean" function for default values of optional
        parameters
        2. Check _CountingAttr object returned by "configurable_boolean" function for specified values of optional
        parameters
        """

        def check_configurable_boolean(
            boolean_instance,
            expected_description: str = "Default configurable boolean description",
            expected_warning: str = None,
            expected_editable: bool = True,
            expected_visible_in_ui: bool = True,
            expected_affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
            expected_ui_rules: UIRules = NullUIRules(),
            expected_auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
            expected_auto_hpo_value: bool = None,
        ):
            expected_metadata = {
                "default_value": True,
                "description": expected_description,
                "header": "configurable_boolean header",
                "warning": expected_warning,
                "editable": expected_editable,
                "visible_in_ui": expected_visible_in_ui,
                "affects_outcome_of": expected_affects_outcome_of,
                "ui_rules": expected_ui_rules,
                "type": ConfigElementType.BOOLEAN,
                "auto_hpo_state": expected_auto_hpo_state,
                "auto_hpo_value": expected_auto_hpo_value,
            }
            assert isinstance(boolean_instance, _make._CountingAttr)
            assert boolean_instance._default
            assert boolean_instance.type == bool
            assert isinstance(boolean_instance._validator, validators._InstanceOfValidator)  # type: ignore
            assert boolean_instance._validator.type == bool
            assert boolean_instance.metadata == expected_metadata

        # Checking _CountingAttr object returned by "configurable_boolean" for default values of optional parameters
        default_value = True
        header = "configurable_boolean header"
        actual_boolean = configurable_boolean(default_value=default_value, header=header)
        check_configurable_boolean(boolean_instance=actual_boolean)  # type: ignore
        # Checking _CountingAttr object returned by "configurable_boolean" for specified values of optional parameters
        description = "configurable_boolean description"
        warning = "configurable_boolean warning"
        editable = False
        visible_in_ui = False
        affects_outcome_of = ModelLifecycle.TESTING
        ui_rules = self.ui_rules
        auto_hpo_state = AutoHPOState.POSSIBLE
        auto_hpo_value = False

        actual_boolean = configurable_boolean(
            default_value=default_value,
            header=header,
            description=description,
            warning=warning,
            editable=editable,
            visible_in_ui=visible_in_ui,
            affects_outcome_of=affects_outcome_of,
            ui_rules=ui_rules,
            auto_hpo_value=auto_hpo_value,
            auto_hpo_state=auto_hpo_state,
        )
        check_configurable_boolean(
            boolean_instance=actual_boolean,  # type: ignore
            expected_description=description,
            expected_warning=warning,
            expected_editable=editable,
            expected_visible_in_ui=visible_in_ui,
            expected_affects_outcome_of=affects_outcome_of,
            expected_ui_rules=ui_rules,
            expected_auto_hpo_value=auto_hpo_value,
            expected_auto_hpo_state=auto_hpo_state,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_float_selectable(self):
        """
        <b>Description:</b>
        Check "float_selectable" function

        <b>Input data:</b>
        "default_value" float, "header" string, "options" list, "description" string, "warning" string, "editable" bool,
        "visible_in_ui" bool, "affects_outcome_of" ModelLifecycle element, "ui_rules" UIRules object

        <b>Expected results:</b>
        Test passes if _CountingAttr object returned by "float_selectable" function is equal to expected

        <b>Steps</b>
        1. Check _CountingAttr object returned by "float_selectable" function for default values of optional parameters
        2. Check _CountingAttr object returned by "float_selectable" function for specified values of optional
        parameters
        """

        def check_float_selectable(
            float_selectable_instance,
            expected_description: str = "Default selectable description",
            expected_warning: str = None,
            expected_editable: bool = True,
            expected_visible_in_ui: bool = True,
            expected_affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
            expected_ui_rules: UIRules = NullUIRules(),
            expected_auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
            expected_auto_hpo_value: float = None,
        ):
            expected_metadata = {
                "default_value": 0.1,
                "description": expected_description,
                "header": "float_selectable header",
                "warning": expected_warning,
                "editable": expected_editable,
                "visible_in_ui": expected_visible_in_ui,
                "affects_outcome_of": expected_affects_outcome_of,
                "ui_rules": expected_ui_rules,
                "type": ConfigElementType.FLOAT_SELECTABLE,
                "options": [0.2, 1.4, 2.8],
                "auto_hpo_state": expected_auto_hpo_state,
                "auto_hpo_value": expected_auto_hpo_value,
            }
            assert isinstance(float_selectable_instance, _make._CountingAttr)
            assert float_selectable_instance._default
            assert float_selectable_instance.type == float
            assert float_selectable_instance.metadata == expected_metadata

        # Checking _CountingAttr object returned by "float_selectable" for default values of optional parameters
        default_value = 0.1
        header = "float_selectable header"
        options = [0.2, 1.4, 2.8]
        actual_float_selectable = float_selectable(default_value=default_value, options=options, header=header)
        check_float_selectable(float_selectable_instance=actual_float_selectable)  # type: ignore
        # Checking _CountingAttr object returned by "float_selectable" for specified values of optional parameters
        description = "float_selectable description"
        warning = "float_selectable warning"
        editable = False
        visible_in_ui = False
        affects_outcome_of = ModelLifecycle.TESTING
        ui_rules = self.ui_rules
        auto_hpo_state = AutoHPOState.POSSIBLE
        auto_hpo_value = options[-1]

        actual_float_selectable = float_selectable(
            default_value=default_value,
            options=options,
            header=header,
            description=description,
            warning=warning,
            editable=editable,
            visible_in_ui=visible_in_ui,
            affects_outcome_of=affects_outcome_of,
            ui_rules=ui_rules,
            auto_hpo_value=auto_hpo_value,
            auto_hpo_state=auto_hpo_state,
        )
        check_float_selectable(
            float_selectable_instance=actual_float_selectable,  # type: ignore
            expected_description=description,
            expected_warning=warning,
            expected_editable=editable,
            expected_visible_in_ui=visible_in_ui,
            expected_affects_outcome_of=affects_outcome_of,
            expected_ui_rules=ui_rules,
            expected_auto_hpo_value=auto_hpo_value,
            expected_auto_hpo_state=auto_hpo_state,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_selectable(self):
        """
        <b>Description:</b>
        Check "selectable" function

        <b>Input data:</b>
        "default_value" ConfigurableEnum element, "header" string, "description" string, "warning" string, "editable"
        bool, "visible_in_ui" bool, "affects_outcome_of" ModelLifecycle element, "ui_rules" UIRules object

        <b>Expected results:</b>
        Test passes if _CountingAttr object returned by "selectable" function is equal to expected

        <b>Steps</b>
        1. Check _CountingAttr object returned by "selectable" function for default values of optional parameters
        2. Check _CountingAttr object returned by "selectable" function for specified values of optional parameters
        """

        def check_selectable(
            selectable_instance,
            expected_description: str = "Default selectable description",
            expected_warning: str = None,
            expected_editable: bool = True,
            expected_visible_in_ui: bool = True,
            expected_affects_outcome_of: ModelLifecycle = ModelLifecycle.NONE,
            expected_ui_rules: UIRules = NullUIRules(),
            expected_auto_hpo_state: AutoHPOState = AutoHPOState.NOT_POSSIBLE,
            expected_auto_hpo_value: SomeEnumSelectable = None,
        ):
            expected_metadata = {
                "default_value": SomeEnumSelectable.OPTION_C,
                "description": expected_description,
                "header": "selectable header",
                "warning": expected_warning,
                "editable": expected_editable,
                "visible_in_ui": expected_visible_in_ui,
                "affects_outcome_of": expected_affects_outcome_of,
                "ui_rules": expected_ui_rules,
                "type": ConfigElementType.SELECTABLE,
                "enum_name": "SomeEnumSelectable",
                "options": {
                    "TEST_NAME1": "test_name_1",
                    "TEST_2": "test_2_test",
                    "BOGUS_NAME": "bogus",
                    "OPTION_C": "option_c",
                },
                "auto_hpo_state": expected_auto_hpo_state,
                "auto_hpo_value": expected_auto_hpo_value,
            }
            assert isinstance(selectable_instance, _make._CountingAttr)
            assert selectable_instance._default == SomeEnumSelectable.OPTION_C
            assert selectable_instance.type == ConfigurableEnum
            assert selectable_instance.metadata == expected_metadata

        # Checking _CountingAttr object returned by "selectable" for default values of optional parameters
        default_value = SomeEnumSelectable.OPTION_C
        header = "selectable header"
        actual_selectable = selectable(default_value=default_value, header=header)
        check_selectable(selectable_instance=actual_selectable)  # type: ignore

        # Checking _CountingAttr object returned by "selectable" for specified values of optional parameters
        description = "selectable description"
        warning = "selectable warning"
        editable = False
        visible_in_ui = False
        affects_outcome_of = ModelLifecycle.TESTING
        ui_rules = self.ui_rules
        auto_hpo_state = AutoHPOState.POSSIBLE
        auto_hpo_value = SomeEnumSelectable.BOGUS_NAME

        actual_selectable = selectable(
            default_value=default_value,
            header=header,
            description=description,
            warning=warning,
            editable=editable,
            visible_in_ui=visible_in_ui,
            affects_outcome_of=affects_outcome_of,
            ui_rules=ui_rules,
            auto_hpo_value=auto_hpo_value,
            auto_hpo_state=auto_hpo_state,
        )
        check_selectable(
            selectable_instance=actual_selectable,  # type: ignore
            expected_description=description,
            expected_warning=warning,
            expected_editable=editable,
            expected_visible_in_ui=visible_in_ui,
            expected_affects_outcome_of=affects_outcome_of,
            expected_ui_rules=ui_rules,
            expected_auto_hpo_state=auto_hpo_state,
            expected_auto_hpo_value=auto_hpo_value,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_string_attribute(self):
        """
        <b>Description:</b>
        Check "string_attribute" function

        <b>Input data:</b>
        "value" string

        <b>Expected results:</b>
        Test passes if _CountingAttr object returned by "string_attribute" function is equal to expected
        """
        value = "some string"
        actual_string = string_attribute(value=value)
        assert isinstance(actual_string, _make._CountingAttr)
        assert actual_string._default == value  # type: ignore
        assert actual_string.kw_only  # type: ignore
        assert actual_string.type == str  # type: ignore
        assert not actual_string._validator  # type: ignore

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_boolean_attribute(self):
        """
        <b>Description:</b>
        Check "boolean_attribute" function

        <b>Input data:</b>
        "value" bool

        <b>Expected results:</b>
        Test passes if _CountingAttr object returned by "boolean_attribute" function is equal to expected
        """
        value = True
        actual_boolean = boolean_attribute(value=value)
        assert isinstance(actual_boolean, _make._CountingAttr)
        assert actual_boolean._default == value  # type: ignore
        assert actual_boolean.kw_only  # type: ignore
        assert actual_boolean.type == bool  # type: ignore
        assert not actual_boolean._validator  # type: ignore



"""Operation module for otx.core.ov.ops.modeuls."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import inspect
from typing import Dict, List, Optional, Union

import torch
from openvino.pyopenvino import Node  # pylint: disable=no-name-in-module

from ..op import Operation
from ..utils import convert_op_to_torch


class OperationModule(torch.nn.Module):
    """OperationModule class."""

    def __init__(
        self,
        op_v: Operation,
        dependent_ops: Union[List[Operation], Dict[str, Optional[Operation]]],
    ):
        super().__init__()

        self.op_v = op_v
        self._dependent_ops = torch.nn.ModuleDict()

        spec = inspect.getfullargspec(op_v.forward)
        kwargs = spec.args[1:]

        self._dependents_with_defaults = []
        if spec.defaults:
            self._dependents_with_defaults = spec.args[-len(spec.defaults) :]

        if isinstance(dependent_ops, list):
            assert len(dependent_ops) == len(kwargs)
            for op_, kwarg in zip(dependent_ops, kwargs):
                self._dependent_ops[kwarg] = op_
        elif isinstance(dependent_ops, dict):
            for kwarg in kwargs:
                self._dependent_ops[kwarg] = dependent_ops[kwarg]
        else:
            raise NotImplementedError

    def forward(self, *args, **kwargs):
        """Operationmodule's forward function."""
        inputs = {k: v() if v is not None else None for k, v in self._dependent_ops.items()}

        if args:
            empty_input_keys = [k for k, v in self._dependent_ops.items() if v is None]
            for key, val in zip(empty_input_keys, args):
                inputs[key] = val
        if kwargs:
            for key, val in kwargs.items():
                if inputs[key] is not None:
                    raise ValueError(f"duplicated key {key}")
                inputs[key] = val

        assert all(v is not None for v in inputs.values() if v not in self._dependents_with_defaults)

        return self.op_v(**inputs)

    @property
    def type(self):  # pylint: disable=invalid-overridden-method
        """Operationmodule's type property."""
        return self.op_v.type

    @property
    def version(self):
        """Operationmodule's version property."""
        return self.op_v.version

    @property
    def name(self):
        """Operationmodule's name property."""
        return self.op_v.name

    @property
    def shape(self):
        """Operationmodule's shape property."""
        return self.op_v.shape

    @property
    def attrs(self):
        """Operationmodule's attrs property."""
        return self.op_v.attrs


def convert_op_to_torch_module(target_op: Node):
    """Convert op Node to torch module."""
    dependent_modules = []
    for in_port in target_op.inputs():
        out_port = in_port.get_source_output()
        parent = out_port.get_node()

        parent_type = parent.get_type_name()
        if parent_type == "Constant":
            dependent_modules.append(convert_op_to_torch(parent))
        else:
            dependent_modules.append(None)
    module = convert_op_to_torch(target_op)
    module = OperationModule(module, dependent_modules)
    return module



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
from networkx.classes.reportviews import (
    DegreeView,
    DiMultiDegreeView,
    EdgeDataView,
    EdgeView,
    InMultiDegreeView,
    NodeView,
    OutMultiDegreeView,
    OutMultiEdgeDataView,
    OutMultiEdgeView,
)
from networkx.exception import NetworkXError, NetworkXNotImplemented

from otx.api.entities.graph import Graph, MultiDiGraph
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestGraph:
    @staticmethod
    def add_edges_to_graph(graph_to_add_edges: Graph, edges_parameters: list) -> None:
        for edge_dict in edges_parameters:
            graph_to_add_edges.add_edge(**edge_dict)

    def non_directed_graph(self) -> Graph:
        non_directed_graph = Graph()
        self.add_edges_to_graph(
            non_directed_graph,
            [
                {"node1": 1, "node2": 2, "edge_value": 2},
                {"node1": 2, "node2": 3, "edge_value": 3},
                {"node1": 3, "node2": 4, "edge_value": 4},
            ],
        )
        return non_directed_graph

    def directed_graph(self) -> Graph:
        directed_graph = Graph(directed=True)
        self.add_edges_to_graph(
            directed_graph,
            [
                {"node1": "A", "node2": "B", "edge_value": 1},
                {"node1": "B", "node2": "C", "edge_value": 2},
                {"node1": "C", "node2": "D", "edge_value": 3},
                {"node1": "D", "node2": "A", "edge_value": 5},
                {"node1": "A", "node2": "D", "edge_value": 4},
            ],
        )
        return directed_graph

    @staticmethod
    def check_graph_non_list_attributes(expected_attributes_dicts: list) -> None:
        for expected_attribute_dict in expected_attributes_dicts:
            assert expected_attribute_dict.get("attribute") == expected_attribute_dict.get("expected_value")

    @staticmethod
    def check_graph_list_attributes(actual_expected_attributes_dict: list) -> None:
        for expected_attribute_dict in actual_expected_attributes_dict:
            attribute = expected_attribute_dict.get("attribute")
            assert isinstance(attribute, expected_attribute_dict.get("expected_type"))
            assert list(attribute) == expected_attribute_dict.get("expected_value")

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @pytest.mark.filterwarnings("ignore::PendingDeprecationWarning")
    def test_graph_initialization(self):
        """
        <b>Description:</b>
        Check Graph class object initialization

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if Graph object "directed" attribute, "edges", "nodes" and "num_labels" properties and "num_nodes"
        method return expected values

        <b>Steps</b>
        1. Check initialization of non-directed Graph object
        2. Check initialization of directed Graph object
        """
        # Checking not directed Graph values
        non_directed_graph = self.non_directed_graph()
        self.check_graph_non_list_attributes(
            [
                {"attribute": non_directed_graph.directed, "expected_value": False},
                {"attribute": non_directed_graph.num_labels, "expected_value": 4},
                {"attribute": non_directed_graph.num_nodes(), "expected_value": 4},
            ]
        )
        self.check_graph_list_attributes(
            [
                {
                    "attribute": non_directed_graph.edges,
                    "expected_type": EdgeDataView,
                    "expected_value": [
                        (1, 2, {"value": 2}),
                        (2, 3, {"value": 3}),
                        (3, 4, {"value": 4}),
                    ],
                },
                {
                    "attribute": non_directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [1, 2, 3, 4],
                },
            ]
        )
        # Checking directed Graph values
        directed_graph = self.directed_graph()
        self.check_graph_non_list_attributes(
            [
                {"attribute": directed_graph.directed, "expected_value": True},
                {"attribute": directed_graph.num_labels, "expected_value": 4},
                {"attribute": directed_graph.num_nodes(), "expected_value": 4},
            ]
        )
        self.check_graph_list_attributes(
            [
                {
                    "attribute": directed_graph.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        ("A", "B", 0, {"value": 1}),
                        ("A", "D", 0, {"value": 4}),
                        ("B", "C", 0, {"value": 2}),
                        ("C", "D", 0, {"value": 3}),
                        ("D", "A", 0, {"value": 5}),
                    ],
                },
                {
                    "attribute": directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": ["A", "B", "C", "D"],
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_get_graph(self):
        """
        <b>Description:</b>
        Check get_graph method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if Graph object returned by get_graph method returns expected attributes

        <b>Steps</b>
        1. Check attributes of Graph returned by get_graph method for non-directed Graph object
        2. Check attributes of Graph returned by get_graph method for directed Graph object
        """
        # Checking non-directed Graph attributes
        non_directed_graph = self.non_directed_graph().get_graph()
        self.check_graph_list_attributes(
            [
                {
                    "attribute": non_directed_graph.degree,
                    "expected_type": DegreeView,
                    "expected_value": [(1, 1), (2, 2), (3, 2), (4, 1)],
                },
                {
                    "attribute": non_directed_graph.edges,
                    "expected_type": EdgeView,
                    "expected_value": [(1, 2), (2, 3), (3, 4)],
                },
                {
                    "attribute": non_directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [1, 2, 3, 4],
                },
            ]
        )
        # Checking directed Graph attributes
        directed_graph = self.directed_graph().get_graph()
        self.check_graph_list_attributes(
            [
                {
                    "attribute": directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": ["A", "B", "C", "D"],
                },
                {
                    "attribute": directed_graph.edges,
                    "expected_type": OutMultiEdgeView,
                    "expected_value": [
                        ("A", "B", 0),
                        ("A", "D", 0),
                        ("B", "C", 0),
                        ("C", "D", 0),
                        ("D", "A", 0),
                    ],
                },
                {
                    "attribute": directed_graph.degree,
                    "expected_type": DiMultiDegreeView,
                    "expected_value": [("A", 3), ("B", 2), ("C", 2), ("D", 3)],
                },
                {
                    "attribute": directed_graph.in_degree,
                    "expected_type": InMultiDegreeView,
                    "expected_value": [("A", 1), ("B", 1), ("C", 1), ("D", 2)],
                },
                {
                    "attribute": directed_graph.out_degree,
                    "expected_type": OutMultiDegreeView,
                    "expected_value": [("A", 2), ("B", 1), ("C", 1), ("D", 1)],
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_set_graph(self):
        """
        <b>Description:</b>
        Check set_graph method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if graph attribute returned after set_graph method has expected values

        <b>Steps</b>
        1. Check graph attribute after setting unequal non-directed graph to non-directed Graph object
        2. Check graph attribute after setting directed graph to non-directed Graph object
        3. Check graph attribute after setting unequal directed graph to directed Graph object
        4. Check graph attribute after setting non-directed graph to directed Graph object
        """
        # Setting unequal non-directed graph to non-directed Graph
        non_directed_graph = self.non_directed_graph()
        unequal_non_directed_graph = Graph()
        self.add_edges_to_graph(
            unequal_non_directed_graph,
            [
                {"node1": 1, "node2": 3, "edge_value": 2},
                {"node1": 2, "node2": 4, "edge_value": 2},
            ],
        )
        non_directed_graph.set_graph(unequal_non_directed_graph)
        assert non_directed_graph.get_graph() == unequal_non_directed_graph
        # Setting directed graph to non-directed Graph
        non_directed_graph = self.non_directed_graph()
        directed_graph = self.directed_graph()
        non_directed_graph.set_graph(directed_graph)
        assert non_directed_graph.get_graph() == directed_graph
        # Setting unequal directed graph to directed Graph
        unequal_directed_graph = Graph(directed=True)
        self.add_edges_to_graph(
            unequal_directed_graph,
            [
                {"node1": "A", "node2": "B", "edge_value": 1},
                {"node1": "B", "node2": "C", "edge_value": 2},
                {"node1": "C", "node2": "B", "edge_value": 1},
            ],
        )
        directed_graph.set_graph(unequal_directed_graph)
        assert directed_graph.get_graph() == unequal_directed_graph
        # Setting non-directed graph to directed Graph
        directed_graph = self.directed_graph()
        non_directed_graph = self.non_directed_graph()
        directed_graph.set_graph(non_directed_graph)
        assert directed_graph.get_graph() == non_directed_graph

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_add_edge(self):
        """
        <b>Description:</b>
        Check add_edge method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if edges attribute returned after add_edge method has expected values

        <b>Steps</b>
        1. Check add_edge method by adding equal edge to non-directed graph (expected, that edges not changed)
        2. Check add_edge method by adding reversed edge to non-directed graph (expected, that edges not changed)
        3. Check add_edge method by adding unequal edge to non-directed graph
        4. Check add_edge method by adding equal edge to directed graph
        5. Check add_edge method by adding reversed edge to directed graph
        6. Check add_edge method by adding unequal edge to directed graph
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        non_directed_graph_edges = [
            (1, 2, {"value": 2}),
            (2, 3, {"value": 3}),
            (3, 4, {"value": 4}),
        ]
        # Adding equal edge to non-directed graph
        non_directed_graph.add_edge(node1=2, node2=3, edge_value=3)
        assert list(non_directed_graph.edges) == non_directed_graph_edges
        # Adding equal reversed edge to non-directed graph
        non_directed_graph.add_edge(node1=3, node2=2, edge_value=3)
        assert list(non_directed_graph.edges) == non_directed_graph_edges
        # Adding unequal edge with existing nodes to non-directed graph
        non_directed_graph.add_edge(node1=1, node2=3, edge_value=1)
        # Adding unequal edge with new node to non-directed graph
        non_directed_graph.add_edge(node1=4, node2=5, edge_value=2)
        assert list(non_directed_graph.edges) == [
            (1, 2, {"value": 2}),
            (1, 3, {"value": 1}),
            (2, 3, {"value": 3}),
            (3, 4, {"value": 4}),
            (4, 5, {"value": 2}),
        ]
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        # Adding equal edge to directed graph
        directed_graph.add_edge(node1="C", node2="D", edge_value=3)
        assert list(directed_graph.edges) == [
            ("A", "B", 0, {"value": 1}),
            ("A", "D", 0, {"value": 4}),
            ("B", "C", 0, {"value": 2}),
            ("C", "D", 0, {"value": 3}),
            ("C", "D", 1, {"value": 3}),
            ("D", "A", 0, {"value": 5}),
        ]
        # Adding equal reversed edge to directed graph
        directed_graph.add_edge(node1="B", node2="A", edge_value=2)
        assert list(directed_graph.edges) == [
            ("A", "B", 0, {"value": 1}),
            ("A", "D", 0, {"value": 4}),
            ("B", "C", 0, {"value": 2}),
            ("B", "A", 0, {"value": 2}),
            ("C", "D", 0, {"value": 3}),
            ("C", "D", 1, {"value": 3}),
            ("D", "A", 0, {"value": 5}),
        ]
        # Adding unequal edge with new nodes to non-directed graph
        directed_graph.add_edge(node1="E", node2="F")
        assert list(directed_graph.edges) == [
            ("A", "B", 0, {"value": 1}),
            ("A", "D", 0, {"value": 4}),
            ("B", "C", 0, {"value": 2}),
            ("B", "A", 0, {"value": 2}),
            ("C", "D", 0, {"value": 3}),
            ("C", "D", 1, {"value": 3}),
            ("D", "A", 0, {"value": 5}),
            ("E", "F", 0, {"value": None}),
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_has_edge_between(self):
        """
        <b>Description:</b>
        Check has_edge_between method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if has_edge_between method returns expected value

        <b>Steps</b>
        1. Check value returned by has_edge_between method for non-directed graph
        2. Check value returned by has_edge_between method for directed graph
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        for node_1, node_2 in ([1, 2], [2, 1], [2, 3], [3, 2], [3, 4], [4, 3]):
            assert non_directed_graph.has_edge_between(node_1, node_2)
        for node_1, node_2 in ([1, 3], [3, 1], [1, 4], [4, 1], [2, 4], [4, 2]):
            assert not non_directed_graph.has_edge_between(node_1, node_2)
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        for node_1, node_2 in (
            ["B", "A"],
            ["C", "B"],
            ["D", "C"],
            ["A", "D"],
            ["D", "A"],
        ):
            assert directed_graph.has_edge_between(node_1, node_2)
        for node_1, node_2 in (
            ["A", "B"],
            ["B", "C"],
            ["C", "D"],
            ["A", "C"],
            ["C", "A"],
            ["B", "D"],
            ["D", "B"],
        ):
            assert not directed_graph.has_edge_between(node_1, node_2)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_neighbors(self):
        """
        <b>Description:</b>
        Check neighbors method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if neighbors method returns expected value

        <b>Steps</b>
        1. Check value returned by neighbors method for non-directed graph
        2. Check value returned by neighbors method for directed graph
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        for node, expected_neighbors in ([1, [2]], [2, [1, 3]], [3, [2, 4]], [4, [3]]):
            assert non_directed_graph.neighbors(node) == expected_neighbors
        assert non_directed_graph.neighbors(5) == []
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        for node, expected_neighbors in (
            ["A", ["B", "D"]],
            ["B", ["C"]],
            ["C", ["D"]],
            ["D", ["A"]],
        ):
            assert directed_graph.neighbors(node) == expected_neighbors
        assert directed_graph.neighbors(1) == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_find_out_edges(self):
        """
        <b>Description:</b>
        Check find_out_edges method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if find_out_edges method returns expected value

        <b>Steps</b>
        1. Check empty list returned by find_out_edges method for non-directed graph
        2. Check value returned by find_out_edges method for directed graph
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        for node in [1, 2, 3, 4]:
            assert non_directed_graph.find_out_edges(node) == []
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        for node, expected_edges in (
            ["A", [("A", "B"), ("A", "D")]],
            ["B", [("B", "C")]],
            ["C", [("C", "D")]],
            ["D", [("D", "A")]],
        ):
            assert list(directed_graph.find_out_edges(node)) == expected_edges
        with pytest.raises(KeyError):
            directed_graph.find_out_edges(1)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_find_in_edges(self):
        """
        <b>Description:</b>
        Check find_in_edges method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if find_in_edges method returns expected value

        <b>Steps</b>
        1. Check empty list returned by find_in_edges method for non-directed graph
        2. Check value returned by find_in_edges method for directed graph
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        for node in [1, 2, 3, 4]:
            assert non_directed_graph.find_in_edges(node) == []
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        for node, expected_edges in (
            ["A", [("D", "A")]],
            ["B", [("A", "B")]],
            ["C", [("B", "C")]],
            ["D", [("C", "D"), ("A", "D")]],
        ):
            assert list(directed_graph.find_in_edges(node)) == expected_edges
        with pytest.raises(KeyError):
            directed_graph.find_in_edges(1)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_find_cliques(self):
        """
        <b>Description:</b>
        Check find_cliques method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if find_cliques method returns generator object with expected values

        <b>Steps</b>
        1. Check value returned by find_cliques generator for non-directed graph
        2. Check NetworkXNotImplemented exception raised when find_cliques method used for directed graph
        """
        # Scenario for non-directed graph
        non_directed_graph_cliques = self.non_directed_graph().find_cliques()
        for expected_clique in ([2, 1], [2, 3], [4, 3]):
            assert next(non_directed_graph_cliques) == expected_clique
        # Scenario for directed graph
        with pytest.raises(NetworkXNotImplemented):
            # this is for networkx<2.6.0
            cliques = self.directed_graph().find_cliques()
            if next(cliques, None):
                # this is for networkx>=2.6.0
                raise NetworkXNotImplemented

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @pytest.mark.filterwarnings("ignore::PendingDeprecationWarning")
    def test_graph_num_labels(self):
        """
        <b>Description:</b>
        Check num_labels property of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if num_labels property returns expected value

        <b>Steps</b>
        1. Check value returned by num_labels property for non-directed graph
        2. Check value returned by num_labels property for directed graph
        """
        for empty_graph in (Graph(), Graph(directed=True)):
            assert empty_graph.num_labels == 0
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        assert non_directed_graph.num_labels == 4
        # Checking num_labels property after adding edge with non existing nodes to non-directed graph
        non_directed_graph.add_edge(5, 6)
        assert non_directed_graph.num_labels == 6
        # Checking num_labels property after removing edge (nodes not removed) from non-directed graph
        non_directed_graph.remove_edges(5, 6)
        assert non_directed_graph.num_labels == 6
        # Checking num_labels property after removing node from non-directed graph
        non_directed_graph.remove_node(1)
        assert non_directed_graph.num_labels == 5
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        assert directed_graph.num_labels == 4
        # Checking num_labels property after adding edge with one existing node to directed graph
        directed_graph.add_edge(node1="A", node2="E", edge_value=7)
        assert directed_graph.num_labels == 5
        # Checking num_labels property after removing edge (nodes not removed) from non-directed graph
        directed_graph.remove_edges(node1="A", node2="B")
        assert directed_graph.num_labels == 5
        # Checking num_labels property after removing node from non-directed graph
        directed_graph.remove_node("E")
        assert directed_graph.num_labels == 4

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_remove_edges(self):
        """
        <b>Description:</b>
        Check remove_edges method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if returned expected value of edges property after using remove_edges method on Graph class object

        <b>Steps</b>
        1. Check value returned by edges property after using remove_edges method by non-directed Graph class object
        2. Check value returned by edges property after using remove_edges method by directed Graph class object
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        # Removing two existing edges from non-directed graph
        non_directed_graph.remove_edges(1, 2)
        non_directed_graph.remove_edges(3, 2)
        assert list(non_directed_graph.edges) == [(3, 4, {"value": 4})]
        # Checking that NetworkXError exception raised when trying to remove non existing edges from non-directed graph
        for node_1, node_2 in ([1, 2], [1, 3], [5, 6]):
            with pytest.raises(NetworkXError):
                non_directed_graph.remove_edges(node_1, node_2)
        # Checking "edges" property of non-directed graph after adding edge which were removed in previous scenario
        non_directed_graph.add_edge(node1=1, node2=2, edge_value=2)
        assert list(non_directed_graph.edges) == [
            (1, 2, {"value": 2}),
            (3, 4, {"value": 4}),
        ]
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        # Removing two existing edges from directed graph, edge "D"-"A" - multi directional
        directed_graph.remove_edges("A", "B")
        directed_graph.remove_edges("D", "A")
        assert list(directed_graph.edges) == [
            ("A", "D", 0, {"value": 4}),
            ("B", "C", 0, {"value": 2}),
            ("C", "D", 0, {"value": 3}),
        ]
        # Checking that NetworkXError exception raised when trying to remove non existing edges from directed graph
        for node_1, node_2 in (["A", "B"], ["A", "C"], ["E", "F"]):
            with pytest.raises(NetworkXError):
                non_directed_graph.remove_edges(node_1, node_2)
        # Checking "edges" property of directed graph after adding edge which were removed in previous scenario
        directed_graph.add_edge(node1="D", node2="A", edge_value=5)
        assert list(directed_graph.edges) == [
            ("A", "D", 0, {"value": 4}),
            ("B", "C", 0, {"value": 2}),
            ("C", "D", 0, {"value": 3}),
            ("D", "A", 0, {"value": 5}),
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_remove_node(self):
        """
        <b>Description:</b>
        Check remove_node method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if returned expected values of nodes property and num_nodes method after using remove_node method on
        Graph class object

        <b>Steps</b>
        1. Check values returned by nodes and edges properties and num_nodes method after using remove_node method by
        non-directed Graph class object
        2. Check values returned by nodes and edges properties and num_nodes method after using remove_node method by
        directed Graph class object
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        # Removing node that connected to one node of non-directed graph
        non_directed_graph.remove_node(1)
        assert non_directed_graph.num_nodes() == 3
        self.check_graph_list_attributes(
            [
                {
                    "attribute": non_directed_graph.edges,
                    "expected_type": EdgeDataView,
                    "expected_value": [(2, 3, {"value": 3}), (3, 4, {"value": 4})],
                },
                {
                    "attribute": non_directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [2, 3, 4],
                },
            ]
        )
        # Checking that NetworkXError exception raised when removing node non existing nodes from non-directed graph
        for node in [1, 5]:
            with pytest.raises(NetworkXError):
                non_directed_graph.remove_node(node)
        # Removing node that connected to two nodes of non-directed graph
        non_directed_graph.remove_node(3)
        assert non_directed_graph.num_nodes() == 2
        self.check_graph_list_attributes(
            [
                {
                    "attribute": non_directed_graph.edges,
                    "expected_type": EdgeDataView,
                    "expected_value": [],
                },
                {
                    "attribute": non_directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [2, 4],
                },
            ]
        )
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        # Removing node that connected to two nodes of non-directed graph and has multi-direction edge
        directed_graph.remove_node("A")
        assert directed_graph.num_nodes() == 3
        self.check_graph_list_attributes(
            [
                {
                    "attribute": directed_graph.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        ("B", "C", 0, {"value": 2}),
                        ("C", "D", 0, {"value": 3}),
                    ],
                },
                {
                    "attribute": directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": ["B", "C", "D"],
                },
            ]
        )
        # Checking that NetworkXError exception raised when removing node non existing nodes from directed graph
        for node in ["A", "E"]:
            with pytest.raises(NetworkXError):
                directed_graph.remove_node(node)
        # Removing node that connected to two nodes of directed graph, it causes removal of all edges
        directed_graph.remove_node("C")
        assert non_directed_graph.num_nodes() == 2
        self.check_graph_list_attributes(
            [
                {
                    "attribute": directed_graph.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [],
                },
                {
                    "attribute": directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": ["B", "D"],
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_num_nodes(self):
        """
        <b>Description:</b>
        Check num_nodes method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if num_nodes method returns expected value

        <b>Steps</b>
        1. Check values returned by num_nodes method for non-directed Graph class object
        2. Check values returned by num_nodes method for directed Graph class object
        """
        for empty_graph in [Graph(), Graph(directed=True)]:
            assert empty_graph.num_nodes() == 0
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        assert non_directed_graph.num_nodes() == 4
        # Checking num_nodes method after adding one node to non-directed graph
        non_directed_graph.add_node(5)
        assert non_directed_graph.num_nodes() == 5
        # Checking num_nodes method after adding edge with one existing node to non-directed graph
        non_directed_graph.add_edge(1, 6)
        assert non_directed_graph.num_nodes() == 6
        # Checking num_nodes method after adding edge with non-existing nodes to non-directed graph
        non_directed_graph.add_edge(7, 8)
        assert non_directed_graph.num_nodes() == 8
        # Checking num_nodes method after removing node from non-directed graph
        non_directed_graph.remove_node(1)
        assert non_directed_graph.num_nodes() == 7
        # Checking num_nodes method after removing edge from non-directed graph
        non_directed_graph.remove_edges(2, 3)
        assert non_directed_graph.num_nodes() == 7
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        assert directed_graph.num_nodes() == 4
        # Checking num_nodes method after adding one node to directed graph
        directed_graph.add_node("E")
        assert directed_graph.num_nodes() == 5
        # Checking num_nodes method after adding edge with one existing node to directed graph
        directed_graph.add_edge("A", "F")
        assert directed_graph.num_nodes() == 6
        # Checking num_nodes method after adding edge with non-existing nodes to directed graph
        directed_graph.add_edge("G", "H")
        assert directed_graph.num_nodes() == 8
        # Checking num_nodes method after removing node from directed graph
        directed_graph.remove_node("A")
        assert directed_graph.num_nodes() == 7
        # Checking num_nodes method after removing edge from non-directed graph
        directed_graph.remove_edges("B", "C")
        assert directed_graph.num_nodes() == 7

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_eq(self):
        """
        <b>Description:</b>
        Check __eq__ method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if __eq__ method returns expected value

        <b>Steps</b>
        1. Check values returned by __eq__ method for non-directed Graph class object
        2. Check values returned by __eq__ method for directed Graph class object
        """
        non_directed_graph = Graph()
        directed_graph = Graph(directed=True)
        unequal_nodes_graph = Graph()
        unequal_edges_graph = Graph()
        equal_non_directed_graph = Graph()
        graph_edges_list = [
            {"node1": 1, "node2": 2, "edge_value": 1},
            {"node1": 2, "node2": 3, "edge_value": 2},
        ]
        for no_edges_graph in [
            non_directed_graph,
            directed_graph,
            equal_non_directed_graph,
            unequal_nodes_graph,
            unequal_edges_graph,
        ]:
            self.add_edges_to_graph(no_edges_graph, graph_edges_list)

        # Checking __eq__ method for equal non-directed graphs
        assert non_directed_graph == equal_non_directed_graph
        # Checking __eq__ method for non-directed graphs with unequal directed attributes
        assert non_directed_graph != directed_graph
        # Checking __eq__ method for non-directed graphs with unequal nodes
        unequal_nodes_graph.add_node(4)
        assert non_directed_graph != unequal_nodes_graph
        # Checking __eq__ method for non-directed graphs with unequal edges
        unequal_edges_graph.add_edge(node1=1, node2=3, edge_value=4)
        assert non_directed_graph != unequal_edges_graph
        # Checking __eq__ method by comparing non-directed Graph with object of other type
        assert non_directed_graph != str
        # Check for non-directed graph
        equal_directed_graph = Graph(directed=True)
        unequal_nodes_graph = Graph(directed=True)
        unequal_edges_graph = Graph(directed=True)
        for no_edges_graph in [
            equal_directed_graph,
            unequal_nodes_graph,
            unequal_edges_graph,
        ]:
            self.add_edges_to_graph(no_edges_graph, graph_edges_list)
        # Checking __eq__ method for equal directed graphs
        assert directed_graph == equal_directed_graph
        # Checking __eq__ method for directed graphs with unequal directed attributes
        assert directed_graph != non_directed_graph
        # Checking __eq__ method for directed graphs with unequal nodes
        unequal_nodes_graph.add_node(4)
        assert directed_graph != unequal_nodes_graph
        # Checking __eq__ method for directed graphs with unequal edges
        unequal_edges_graph.add_edge(2, 1, 4)
        assert directed_graph != unequal_edges_graph
        # Checking __eq__ method by comparing directed Graph with object of other type
        assert directed_graph != str

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_graph_descendants(self):
        """
        <b>Description:</b>
        Check descendants method of Graph class object

        <b>Input data:</b>
        Graph objects with specified "directed" and "edges" parameters

        <b>Expected results:</b>
        Test passes if descendants method returns expected value

        <b>Steps</b>
        1. Check values returned by descendants method for non-directed Graph class object
        2. Check values returned by descendants method for directed Graph class object
        """
        # Scenario for non-directed graph
        non_directed_graph = self.non_directed_graph()
        assert non_directed_graph.descendants(1) == [1, 2, 3]
        assert non_directed_graph.descendants(2) == [2, 2, 3]
        assert non_directed_graph.descendants(3) == [3, 2, 3]
        assert non_directed_graph.descendants(4) == [4, 3, 2]
        # Checking descendants after removing node from non-directed graph
        non_directed_graph.remove_node(2)
        assert non_directed_graph.descendants(1) == []
        assert non_directed_graph.descendants(2) == []
        assert non_directed_graph.descendants(3) == [3]
        assert non_directed_graph.descendants(4) == [4]
        # Scenario for directed graph
        directed_graph = self.directed_graph()
        assert directed_graph.descendants("A") == ["D", "C", "B", "A", "A"]
        assert directed_graph.descendants("B") == ["A", "D", "C", "B", "A"]
        assert directed_graph.descendants("C") == ["B", "A", "D", "C", "A"]
        assert directed_graph.descendants("D") == ["C", "B", "A", "D", "A"]
        # Checking descendants after removing node from directed graph
        directed_graph.remove_node("B")
        assert directed_graph.descendants("A") == ["D", "C", "A"]
        assert directed_graph.descendants("B") == []
        assert directed_graph.descendants("C") == []
        assert directed_graph.descendants("D") == ["C", "A", "D"]


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestMultiDiGraph:
    @staticmethod
    def multi_di_graph():
        multi_di_graph = MultiDiGraph()
        TestGraph.add_edges_to_graph(
            multi_di_graph,
            [
                {"node1": (1, 1), "node2": (1, 2)},
                {"node1": (1, 2), "node2": (3, 1)},
                {"node1": (3, 1), "node2": (2, 1)},
            ],
        )
        return multi_di_graph

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @pytest.mark.filterwarnings("ignore::PendingDeprecationWarning")
    def test_multi_di_graph(self):
        """
        <b>Description:</b>
        Check MultiDiGraph class object initialization

        <b>Input data:</b>
        MultiDiGraph object with specified "edges" parameter

        <b>Expected results:</b>
        Test passes if MultiDiGraph object "directed" attribute, "edges", "nodes" and "num_labels" properties and
        "num_nodes" method return expected values
        """
        multi_di_graph = self.multi_di_graph()
        TestGraph.check_graph_non_list_attributes(
            [
                {"attribute": multi_di_graph.directed, "expected_value": True},
                {"attribute": multi_di_graph.num_labels, "expected_value": 4},
                {"attribute": multi_di_graph.num_nodes(), "expected_value": 4},
            ]
        )
        TestGraph.check_graph_list_attributes(
            [
                {
                    "attribute": multi_di_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [(1, 1), (1, 2), (3, 1), (2, 1)],
                },
                {
                    "attribute": multi_di_graph.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        ((1, 1), (1, 2), 0, {"value": None}),
                        ((1, 2), (3, 1), 0, {"value": None}),
                        ((3, 1), (2, 1), 0, {"value": None}),
                    ],
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_multi_di_graph_topological_sort(self):
        """
        <b>Description:</b>
        Check topological_sort method of MultiDiGraph class object

        <b>Input data:</b>
        MultiDiGraph objects with specified "edges" parameter

        <b>Expected results:</b>
        Test passes if topological_sort method returns generator object with expected values
        """
        multi_di_graph = self.multi_di_graph()
        topological_sort = multi_di_graph.topological_sort()
        for expected_value in [(1, 1), (1, 2), (3, 1), (2, 1)]:
            assert next(topological_sort) == expected_value



"""Collections of Utils for HPO."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Literal, Optional


def left_vlaue_is_better(val1, val2, mode: Literal["max", "min"]) -> bool:
    """Check left value is better than right value.

    Whether check it's greather or lesser is changed depending on 'model'.

    Args:
        val1 : value to check that it's bigger than other value.
        val2 : value to check that it's bigger than other value.
        mode (Literal['max', 'min']): value to decide whether better means greater or lesser.

    Returns:
        bool: whether val1 is better than val2.
    """
    check_mode_input(mode)
    if mode == "max":
        return val1 > val2
    return val1 < val2


def check_positive(value, variable_name: Optional[str] = None, error_message: Optional[str] = None):
    """Validate that value is positivle.

    Args:
        value (Any): value to validate.
        variable_name (Optional[str], optional): name of value. It's used for error message. Defaults to None.
        error_message (Optional[str], optional): Error message to use when type is different. Defaults to None.

    Raises:
        ValueError: If value isn't positive, the error is raised.
    """
    if value <= 0:
        if error_message is not None:
            message = error_message
        elif variable_name:
            message = f"{variable_name} should be positive.\n" f"your value : {value}"
        else:
            raise ValueError
        raise ValueError(message)


def check_not_negative(value, variable_name: Optional[str] = None, error_message: Optional[str] = None):
    """Validate that value isn't negative.

    Args:
        value (Any): value to validate.
        variable_name (Optional[str], optional): name of value. It's used for error message. Defaults to None.
        error_message (Optional[str], optional): Error message to use when type is different. Defaults to None.

    Raises:
        ValueError: If value is negative, the error is raised.
    """
    if value < 0:
        if error_message is not None:
            message = error_message
        elif variable_name:
            message = f"{variable_name} should be positive.\n" f"your value : {value}"
        else:
            raise ValueError
        raise ValueError(message)


def check_mode_input(mode: str):
    """Validate that mode is 'max' or 'min'.

    Args:
        mode (str): string to validate.

    Raises:
        ValueError: If 'mode' is not both 'max' and 'min', the error is raised.
    """
    if mode not in ["max", "min"]:
        raise ValueError("mode should be max or min.\n" f"Your value : {mode}")



"""This module tests classes related to metadata"""

# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import re

import pytest

from otx.api.entities.metadata import (
    FloatMetadata,
    FloatType,
    IMetadata,
    MetadataItemEntity,
)
from otx.api.entities.model import ModelEntity
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIMetadata:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_imetadata(self):
        """
        <b>Description:</b>
        To test IMetadata class

        <b>Input data:</b>
        Initialized instance of IMetadata class

        <b>Expected results:</b>
        1. Initialized instance is instance of the class
        2. Default value of field has expected value:  "typing.Union[str, NoneType]"
        3. Changed fields value has expected value: "String"

        <b>Steps</b>
        1. Create IMetadata
        2. Check default value of class field
        3. Change value of the field
        """

        test_instance = IMetadata()
        assert isinstance(test_instance, IMetadata)
        assert str(test_instance.name) in ["typing.Union[str, NoneType]", "typing.Optional[str]"]

        test_instance.name = "String"
        assert test_instance.name == "String"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestFloatType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_float_type_members(self):
        """
        <b>Description:</b>
        To test FloatType enumeration members

        <b>Input data:</b>
        Initialized instance of FloatType enum class

        <b>Expected results:</b>
        1. Enum members return correct values:
            FLOAT = 1
            EMBEDDING_VALUE = 2
            ACTIVE_SCORE = 3
        2. In case incorrect member it raises AttributeError exception
        3. In case incorrect member value it raises ValueError exception

        <b>Steps</b>
        0. Create FloatType
        1. Check members
        2. Check incorrect member
        3. Check incorrect member value
        """

        test_instance = FloatType

        for i in range(1, 4):
            assert test_instance(i) in list(FloatType)

        with pytest.raises(AttributeError):
            test_instance.WRONG

        with pytest.raises(ValueError):
            test_instance(6)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_float_type_magic_str(self):
        """
        <b>Description:</b>
        To test FloatType __str__ method

        <b>Input data:</b>
        Initialized instance of FloatType enum

        <b>Expected results:</b>
        1. __str__ returns correct string for every enum member
        2. In case incorrect member it raises AttributeError exception
        3. In case incorrect member value it raises ValueError exception

        <b>Steps</b>
        0. Create FloatType
        1. Check returning value of __str__ method
        2. Try incorrect field name
        3. Try incorrect value name
        """
        test_instance = FloatType
        magic_str_list = [str(i) for i in list(FloatType)]

        for i in range(1, 4):
            assert str(test_instance(i)) in magic_str_list

        with pytest.raises(AttributeError):
            str(test_instance.WRONG)

        with pytest.raises(ValueError):
            str(test_instance(6))


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestMetadataItemEntity:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_metadata_item_entity(self):
        """
        <b>Description:</b>
        To test MetadataItemEntity class

        <b>Input data:</b>
        Initialized instances of IMetadata, ModelEntity and MetadataItemEntity classes

        <b>Expected results:</b>
        1. Initialized instances
        2. Default values of fields are None
        3. repr method returns expected value "NullMetadata(None, None)" after initiation for both instances
        4. Instance test_instance1 fields values changed and have expected values:
            name == String1
            value == 0
        5. repr method of test_instance1 returns expected value "NullMetadata(String1, 0)"
        6. == method behavior is expected

        <b>Steps</b>
        1. Create class instances test_instance0 and test_instance1
        2. Perform checks of its field default values after initiations
        3. Perform checks of repr method returns expected values after initiations
        4. Change fields value of test_instance1
        5. Perform checks of repr method returns expected values after changes
        6. Check that test_instance0 == test_instance1
        """
        i_metadata = IMetadata()
        i_metadata.name = "default_i_metadata"
        test_data0 = test_data1 = i_metadata.name
        i_metadata.name = "i_metadata"
        test_data2 = i_metadata.name
        test_model0 = test_model1 = ModelEntity(train_dataset="default_dataset", configuration="default_config")
        test_instance0 = MetadataItemEntity(test_data0, test_model0)
        test_instance1 = MetadataItemEntity(test_data1, test_model1)
        test_instance2 = MetadataItemEntity(test_data2, test_model1)
        assert test_instance0 == test_instance1 != test_instance2
        __repr = repr(test_instance0)
        repr_pattern = (
            r"MetadataItemEntity\(model=\<otx.api.entities.model.ModelEntity object at" r" 0x[a-fA-F0-9]{10,32}\>\)"
        )
        assert re.match(repr_pattern, __repr)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestFloatMetadata:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_float_metadata(self):
        """
        <b>Description:</b>
        To test FloatMetadata class

        <b>Input data:</b>
        Initialized instance of FloatMetadata

        <b>Expected results:</b>
        1. It raises TypeError in case attempt of initiation with wrong parameters numbers
        2. Fields of instances are with correct values
        3. repr method returns correct strings then used against each instance
        4. '==' method works as expected

        <b>Steps</b>
        1. Attempt to initiate class instance with wrong parameters numbers
        2. Initiate three class instances:
            two of them with similar set of init values, third one with different one.
        3. Check repr method
        4. Check __eq__ method
        """
        with pytest.raises(TypeError):
            FloatMetadata()

        with pytest.raises(TypeError):
            FloatMetadata("only name")

        test_inst0 = FloatMetadata(name="Instance0", value=42)
        assert test_inst0.name == "Instance0"
        assert test_inst0.value == 42
        assert repr(test_inst0.float_type) == "<FloatType.FLOAT: 1>"
        assert repr(test_inst0) == "FloatMetadata(Instance0, 42, FLOAT)"

        test_inst1 = FloatMetadata(name="Instance1", value=42.0)
        assert test_inst1.name == "Instance1"
        assert test_inst1.value == 42.0
        assert repr(test_inst1.float_type) == "<FloatType.FLOAT: 1>"
        assert repr(test_inst1) == "FloatMetadata(Instance1, 42.0, FLOAT)"

        test_inst2 = FloatMetadata(name="Instance0", value=42)
        assert test_inst2.name == "Instance0"
        assert test_inst2.value == 42
        assert repr(test_inst2.float_type) == "<FloatType.FLOAT: 1>"
        assert repr(test_inst2) == "FloatMetadata(Instance0, 42, FLOAT)"

        assert test_inst0 == test_inst2 != test_inst1



import pytest

from otx.hpo.resource_manager import (
    CPUResourceManager,
    GPUResourceManager,
    _remove_none_from_dict,
    get_resource_manager,
)
from tests.test_suite.e2e_test_system import e2e_pytest_component


@pytest.fixture
def cpu_resource_manager():
    return CPUResourceManager(num_parallel_trial=4)


@pytest.fixture
def gpu_resource_manager():
    return GPUResourceManager(num_gpu_for_single_trial=1, available_gpu="0,1,2,3")


class TestCPUResourceManager:
    @e2e_pytest_component
    @pytest.mark.parametrize("num_parallel_trial", [1, 5, 10])
    def test_init(self, num_parallel_trial):
        CPUResourceManager(num_parallel_trial)

    @e2e_pytest_component
    @pytest.mark.parametrize("num_parallel_trial", [-1, 0])
    def test_init_with_not_positive_num_parallel_trial(self, num_parallel_trial):
        with pytest.raises(ValueError):
            CPUResourceManager(num_parallel_trial)

    @e2e_pytest_component
    def test_reserve_resource(self, cpu_resource_manager):
        num_parallel_trial = cpu_resource_manager._num_parallel_trial

        for i in range(num_parallel_trial):
            assert cpu_resource_manager.reserve_resource(i) == {}

        for i in range(10):
            assert cpu_resource_manager.reserve_resource(i) is None

    @e2e_pytest_component
    def test_reserve_resource_reserved_already(self, cpu_resource_manager):
        cpu_resource_manager.reserve_resource(0)
        with pytest.raises(RuntimeError):
            cpu_resource_manager.reserve_resource(0)

    @e2e_pytest_component
    def test_release_resource(self, cpu_resource_manager):
        cpu_resource_manager.reserve_resource(1)
        cpu_resource_manager.release_resource(1)

    @e2e_pytest_component
    def test_release_unreserved_resource(self, cpu_resource_manager):
        cpu_resource_manager.release_resource(1)

    @e2e_pytest_component
    def test_have_available_resource(self, cpu_resource_manager):
        num_parallel_trial = cpu_resource_manager._num_parallel_trial

        for i in range(num_parallel_trial):
            assert cpu_resource_manager.have_available_resource()
            cpu_resource_manager.reserve_resource(i)

        assert not cpu_resource_manager.have_available_resource()


class TestGPUResourceManager:
    @e2e_pytest_component
    def test_init(self):
        GPUResourceManager(num_gpu_for_single_trial=1, available_gpu="0,1,2")

    @e2e_pytest_component
    @pytest.mark.parametrize("num_gpu_for_single_trial", [-1, 0])
    def test_init_not_positive_num_gpu(self, num_gpu_for_single_trial):
        with pytest.raises(ValueError):
            GPUResourceManager(num_gpu_for_single_trial=num_gpu_for_single_trial)

    @e2e_pytest_component
    @pytest.mark.parametrize("available_gpu", [",", "a,b", "0,a", ""])
    def test_init_wrong_available_gpu_value(self, available_gpu):
        with pytest.raises(ValueError):
            GPUResourceManager(available_gpu=available_gpu)

    @e2e_pytest_component
    def test_reserve_resource(self):
        num_gpu_for_single_trial = 2
        num_gpus = 8
        max_parallel = num_gpus // num_gpu_for_single_trial
        gpu_resource_manager = GPUResourceManager(
            num_gpu_for_single_trial=num_gpu_for_single_trial,
            available_gpu=",".join([str(val) for val in range(num_gpus)]),
        )
        num_gpus = len(gpu_resource_manager._available_gpu)

        for i in range(max_parallel):
            env = gpu_resource_manager.reserve_resource(i)
            assert env is not None
            assert "CUDA_VISIBLE_DEVICES" in env
            assert len(env["CUDA_VISIBLE_DEVICES"].split(",")) == num_gpu_for_single_trial

        for i in range(max_parallel, max_parallel + 10):
            assert gpu_resource_manager.reserve_resource(i) is None

    @e2e_pytest_component
    def test_reserve_resource_reserved_already(self, gpu_resource_manager):
        gpu_resource_manager.reserve_resource(0)
        with pytest.raises(RuntimeError):
            gpu_resource_manager.reserve_resource(0)

    @e2e_pytest_component
    def test_release_resource(self, gpu_resource_manager):
        gpu_resource_manager.reserve_resource(1)
        gpu_resource_manager.release_resource(1)

    @e2e_pytest_component
    def test_release_unreserved_resource(self, gpu_resource_manager):
        gpu_resource_manager.release_resource(1)

    @e2e_pytest_component
    def test_have_available_resource(self):
        num_gpu_for_single_trial = 2
        num_gpus = 8
        max_parallel = num_gpus // num_gpu_for_single_trial
        gpu_resource_manager = GPUResourceManager(
            num_gpu_for_single_trial=num_gpu_for_single_trial,
            available_gpu=",".join([str(val) for val in range(num_gpus)]),
        )
        num_gpus = len(gpu_resource_manager._available_gpu)

        for i in range(max_parallel):
            assert gpu_resource_manager.have_available_resource()
            gpu_resource_manager.reserve_resource(i)

        for i in range(max_parallel, max_parallel + 10):
            assert not gpu_resource_manager.have_available_resource()


@e2e_pytest_component
def test_get_resource_manager_cpu():
    manager = get_resource_manager(resource_type="cpu", num_parallel_trial=4)
    assert isinstance(manager, CPUResourceManager)


@e2e_pytest_component
def test_get_resource_manager_gpu():
    num_gpu_for_single_trial = 1
    available_gpu = "0,1,2,3"
    manager = get_resource_manager(
        resource_type="gpu", num_gpu_for_single_trial=num_gpu_for_single_trial, available_gpu=available_gpu
    )
    assert isinstance(manager, GPUResourceManager)


@e2e_pytest_component
def test_get_resource_manager_wrong_resource_type():
    with pytest.raises(ValueError):
        get_resource_manager("wrong")


@e2e_pytest_component
def test_get_resource_manager_gpu_without_available_gpu(mocker):
    mock_is_available = mocker.patch("otx.hpo.resource_manager.torch.cuda.is_available")
    mock_is_available.return_value = False

    manager = get_resource_manager("gpu")
    assert isinstance(manager, CPUResourceManager)


@e2e_pytest_component
def test_remove_none_from_dict():
    some_dict = {"a": 1, "b": None}
    ret = _remove_none_from_dict(some_dict)
    assert ret == {"a": 1}



"""Module for defining distance utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import torch.distributed as dist


def get_dist_info():  # pylint: disable=inconsistent-return-statements
    """A function that retrieves information about the current distributed training environment."""
    if dist.is_available():
        # data distributed parallel
        try:
            return dist.get_rank(), dist.get_world_size(), True
        except RuntimeError:
            return 0, 1, False



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.api.entities.url import URL
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestURL:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_repo(self):
        """
        <b>Description:</b>
        Check that URL can correctly decode a given URL

        <b>Input data:</b>
        Some URL strings

        <b>Expected results:</b>
        Test passes if the correct scheme and path can be retrieved from the url

        <b>Steps</b>
        1. Create URL instances
        2. Check scheme of URL
        3. Check path of URL
        """
        test_url = URL("binaryrepo:/images/file_%20_whatever.jpg")
        test_url2 = URL("binaryrepo://intel.com/images/file_%20_whatever.jpg?blaat=none")

        assert test_url.scheme == "binaryrepo"
        assert test_url2.scheme == "binaryrepo"

        assert test_url.path == "/images/file_%20_whatever.jpg"
        assert test_url2.path == "/images/file_%20_whatever.jpg"

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_url_initialization(self):
        """
        <b>Description:</b>
        Check initialized URL object attributes

        <b>Input data:</b>
        Initialized URL object

        <b>Expected results:</b>
        Test passes if attributes of initialized URL object are equal to expected

        <b>Steps</b>
        1. Check attributes of URL object initialized without special characters
        2. Check attributes of URL object initialized with "#" special character
        3. Check attributes of URL object initialized with "?" special character
        4. Check attributes of URL object initialized with netloc
        5. Check attributes of URL object initialized with unsupported character in URL-scheme
        6. Check attributes of URL object initialized without URL-scheme
        7. Check that ValueError exception is raised when initializing URL-class object with incorrect netloc
        """

        def check_url(
            url: URL,
            expected_path: str = "",
            expected_extension: str = "",
            expected_host: str = "",
            expected_scheme: str = "",
            expected_relative_path: str = "",
        ):
            assert url.path == expected_path
            assert url.extension == expected_extension
            assert url.host == expected_host
            assert url.scheme == expected_scheme
            if expected_relative_path == "":
                expected_relative_path = expected_path[1:]
            assert url.relative_path == expected_relative_path

        # Checking attributes of URL initialized without special characters
        check_url(
            url=URL("binaryrepo:/images/file_%20_whatever.jpg"),
            expected_path="/images/file_%20_whatever.jpg",
            expected_extension="jpg",
            expected_scheme="binaryrepo",
        )
        # Checking attributes of URL initialized with "#" special character
        check_url(
            url=URL("binaryrepo:/images/file_%20_whatever#.jpg"),
            expected_path="/images/file_%20_whatever",
            expected_scheme="binaryrepo",
        )
        # Checking attributes of URL initialized with "?" special character
        check_url(
            url=URL("binaryrepo:/i?mages/file_%20_whatever.jpg?some_text.png"),
            expected_path="/i",
            expected_scheme="binaryrepo",
        )
        # Checking attributes of URL initialized with netloc
        check_url(
            url=URL("https://www.some_host.com/images/file_%20_whatever.png"),
            expected_path="/images/file_%20_whatever.png",
            expected_extension="png",
            expected_host="www.some_host.com",
            expected_scheme="https",
        )
        # Checking attributes of URL initialized with unsupported character in URL-scheme
        check_url(
            url=URL("binary_repo:/images/file_%20_whatever.jpg"),
            expected_path="binary_repo:/images/file_%20_whatever.jpg",
            expected_extension="jpg",
            expected_relative_path="binary_repo:/images/file_%20_whatever.jpg",
        )
        # Checking attributes of URL initialized without URL-scheme
        check_url(
            url=URL("images/file_%20_whatever.jpg"),
            expected_path="images/file_%20_whatever.jpg",
            expected_extension="jpg",
            expected_relative_path="images/file_%20_whatever.jpg",
        )
        # Checking that ValueError exception is raised when initializing URL with incorrect netloc
        for incorrect_netloc in [
            "https://[www.some_host.com/images/file_%20_whatever.jpg?some_text.png",
            "https://www.some_host.com]/images/file_%20_whatever.jpg?some_text.png",
        ]:
            with pytest.raises(ValueError):
                URL(incorrect_netloc)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_url_str(self):
        """
        <b>Description:</b>
        Check URL class __str__ method

        <b>Input data:</b>
        Initialized URL object

        <b>Expected results:</b>
        Test passes if string returned by __str__ method is equal to expected

        <b>Steps</b>
        1. Check string returned by __str__ method for URL object initialized with non-empty "scheme" and "host"
        attributes
        2. Check string returned by __str__ method for URL object initialized with non-empty "scheme" and empty "host"
        attributes
        3. Check string returned by __str__ method for URL object initialized with non-empty "host" and empty "scheme"
        attributes
        """
        # Checking value returned by __str__ for URL initialized with non-empty "scheme" and "host"
        assert str(URL("https://www.some_host.com/images/file_%20_whatever.png")) == (
            "https://www.some_host.com/images/file_%20_whatever.png"
        )
        # Checking value returned by __str__ for URL initialized with non-empty "scheme" and empty "host"
        assert str(URL("binaryrepo:/images/file_%20_whatever.jpg")) == "binaryrepo:/images/file_%20_whatever.jpg"
        # Checking value returned by __str__ for URL initialized with non-empty "host" and empty "scheme"
        assert str(URL("//www.some_host.com/images/file_%20_whatever.jpg")) == "/images/file_%20_whatever.jpg"

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_url_repr(self):
        """
        <b>Description:</b>
        Check URL class __repr__ method

        <b>Input data:</b>
        Initialized URL object

        <b>Expected results:</b>
        Test passes if string returned by __repr__ method is equal to expected
        """
        assert repr(URL("https://www.some_host.com/images/file_%20_whatever.png")) == (
            "URL(https://www.some_host.com/images/file_%20_whatever.png)"
        )
        assert repr(URL("binaryrepo:/i?mages/file_%20_whatever.jpg?some_text.png")) == (
            "URL(binaryrepo:/i?mages/file_%20_whatever.jpg?some_text.png)"
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_url_hash(self):
        """
        <b>Description:</b>
        Check URL class __hash__ method

        <b>Input data:</b>
        Initialized URL object

        <b>Expected results:</b>
        Test passes if value returned by __hash__ method is equal to expected
        """
        url = URL("https://www.some_host.com/images/file_%20_whatever.png")
        assert hash(url) == hash(str(url))

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_url_length(self):
        """
        <b>Description:</b>
        Check URL class __len__ method

        <b>Input data:</b>
        Initialized URL object

        <b>Expected results:</b>
        Test passes if value returned by __len__ method is equal to expected
        """
        assert len(URL("https://www.some_host.com/images/file_%20_whatever.png")) == 54
        assert len(URL("binaryrepo:/i?mages/file_%20_whatever.jpg?some_text.png")) == 55

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_url_eq(self):
        """
        <b>Description:</b>
        Check URL class __eq__ method

        <b>Input data:</b>
        Initialized URL object

        <b>Expected results:</b>
        Test passes if value returned by __eq__ method is equal to expected

        1. Check value returned by __eq__ method for URL objects with equal "scheme", "host", and "path" properties
        2. Check value returned by __eq__ method for URL objects with unequal "scheme" properties
        3. Check value returned by __eq__ method for URL objects with unequal "host" properties
        4. Check value returned by __eq__ method for URL objects with unequal "path" properties
        """
        # Checking value returned by __eq__ for URL with equal "scheme", "host", and "path"
        url = URL("https://www.some_host.com/images/file_%20_whatever.png")
        equal_url = URL("https://www.some_host.com/images/file_%20_whatever.png")
        assert url == equal_url
        # Checking value returned by __eq__ for URL with unequal "scheme"
        unequal_url = URL("http://www.some_host.com/images/file_%20_whatever.png")
        assert not url == unequal_url
        # Checking value returned by __eq__ for URL with unequal "host"
        unequal_url = URL("https://www.unequal_host.com/images/file_%20_whatever.png")
        assert not url == unequal_url
        # Checking value returned by __eq__ for URL with unequal "path"
        unequal_url = URL("https://www.some_host.com/images/file_%20_whatever.jpg")
        assert not url == unequal_url



"""OTX Adapters for action recognition backbones - mmaction2."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .movinet_head import MoViNetHead
from .roi_head import AVARoIHead

__all__ = ["AVARoIHead", "MoViNetHead"]
